#!/usr/bin/env python
# License: GPLv3 Copyright: 2011, Kovid Goyal <kovid@kovidgoyal.net>

import datetime
import os
import unittest
from io import BytesIO
from time import time

from calibre.db.tests.base import BaseTest
from calibre.utils.date import utc_tz
from calibre.utils.localization import calibre_langcode_to_name


def p(x):
    return datetime.datetime.strptime(x, '%Y-%m-%d').replace(tzinfo=utc_tz)


class ReadingTest(BaseTest):
    def test_read(self):  # {{{
        "Test the reading of data from the database"
        cache = self.init_cache(self.library_path)
        tests = {
            3: {
                'title': 'Unknown',
                'sort': 'Unknown',
                'authors': ('Unknown',),
                'author_sort': 'Unknown',
                'series': None,
                'series_index': 1.0,
                'rating': None,
                'tags': (),
                'formats': (),
                'identifiers': {},
                'timestamp': datetime.datetime(2011, 9, 7, 19, 54, 41, tzinfo=utc_tz),
                'pubdate': datetime.datetime(2011, 9, 7, 19, 54, 41, tzinfo=utc_tz),
                'last_modified': datetime.datetime(2011, 9, 7, 19, 54, 41, tzinfo=utc_tz),
                'publisher': None,
                'languages': (),
                'comments': None,
                '#enum': None,
                '#authors': (),
                '#date': None,
                '#rating': None,
                '#series': None,
                '#series_index': None,
                '#tags': (),
                '#yesno': None,
                '#comments': None,
                'size': None,
            },
            2: {
                'title': 'Title One',
                'sort': 'One',
                'authors': ('Author One',),
                'author_sort': 'One, Author',
                'series': 'A Series One',
                'series_index': 1.0,
                'tags': ('Tag One', 'Tag Two'),
                'formats': ('FMT1',),
                'rating': 4.0,
                'identifiers': {'test': 'one'},
                'timestamp': datetime.datetime(2011, 9, 5, 21, 6, tzinfo=utc_tz),
                'pubdate': datetime.datetime(2011, 9, 5, 21, 6, tzinfo=utc_tz),
                'publisher': 'Publisher One',
                'languages': ('eng',),
                'comments': '<p>Comments One</p>',
                '#enum': 'One',
                '#authors': ('Custom One', 'Custom Two'),
                '#date': datetime.datetime(2011, 9, 5, 6, 0, tzinfo=utc_tz),
                '#rating': 2.0,
                '#series': 'My Series One',
                '#series_index': 1.0,
                '#tags': ('My Tag One', 'My Tag Two'),
                '#yesno': True,
                '#comments': '<div>My Comments One<p></p></div>',
                'size': 9,
            },
            1: {
                'title': 'Title Two',
                'sort': 'Title Two',
                'authors': ('Author Two', 'Author One'),
                'author_sort': 'Two, Author & One, Author',
                'series': 'A Series One',
                'series_index': 2.0,
                'rating': 6.0,
                'tags': ('Tag One', 'News'),
                'formats': ('FMT1', 'FMT2'),
                'identifiers': {'test': 'two'},
                'timestamp': datetime.datetime(2011, 9, 6, 6, 0, tzinfo=utc_tz),
                'pubdate': datetime.datetime(2011, 8, 5, 6, 0, tzinfo=utc_tz),
                'publisher': 'Publisher Two',
                'languages': ('deu',),
                'comments': '<p>Comments Two</p>',
                '#enum': 'Two',
                '#authors': ('My Author Two',),
                '#date': datetime.datetime(2011, 9, 1, 6, 0, tzinfo=utc_tz),
                '#rating': 4.0,
                '#series': 'My Series Two',
                '#series_index': 3.0,
                '#tags': ('My Tag Two',),
                '#yesno': False,
                '#comments': '<div>My Comments Two<p></p></div>',
                'size': 9,
            },
        }
        for book_id, test in tests.items():
            for field, expected_val in test.items():
                val = cache.field_for(field, book_id)
                if isinstance(val, tuple) and 'authors' not in field and 'languages' not in field:
                    val, expected_val = set(val), set(expected_val)
                self.assertEqual(expected_val, val, f'Book id: {book_id} Field: {field} failed: {expected_val!r} != {val!r}')
        # }}}

    def test_sorting(self):  # {{{
        "Test sorting"
        cache = self.init_cache()
        ae = self.assertEqual

        lmap = {x: cache.field_for('languages', x) for x in (1, 2, 3)}
        lq = sorted(lmap, key=lambda x: calibre_langcode_to_name((lmap[x] or ('',))[0]))
        for field, order in {
            'title': [2, 1, 3],
            'authors': [2, 1, 3],
            'series': [3, 1, 2],
            'tags': [3, 1, 2],
            'rating': [3, 2, 1],
            # 'identifiers': [3, 2, 1], There is no stable sort since 1 and
            # 2 have the same identifier keys
            # 'last_modified': [3, 2, 1], There is no stable sort as two
            # records have the exact same value
            'timestamp': [2, 1, 3],
            'pubdate': [1, 2, 3],
            'publisher': [3, 2, 1],
            'languages': lq,
            'comments': [3, 2, 1],
            '#enum': [3, 2, 1],
            '#authors': [3, 2, 1],
            '#date': [3, 1, 2],
            '#rating': [3, 2, 1],
            '#series': [3, 2, 1],
            '#tags': [3, 2, 1],
            '#yesno': [2, 1, 3],
            '#comments': [3, 2, 1],
            'id': [1, 2, 3],
        }.items():
            x = list(reversed(order))
            ae(order, cache.multisort([(field, True)], ids_to_sort=x), f'Ascending sort of {field} failed')
            ae(x, cache.multisort([(field, False)], ids_to_sort=order), f'Descending sort of {field} failed')

        # Test sorting of is_multiple fields.

        # Author like fields should be sorted by generating sort names from the
        # actual values in entry order
        for field in ('authors', '#authors'):
            ae(
                cache.set_field(field, {1: ('aa bb', 'bb cc', 'cc dd'), 2: ('bb aa', 'xx yy'), 3: ('aa bb', 'bb aa')}),
                {1, 2, 3},
            )
            ae([2, 3, 1], cache.multisort([(field, True)], ids_to_sort=(1, 2, 3)))
            ae([1, 3, 2], cache.multisort([(field, False)], ids_to_sort=(1, 2, 3)))

        # All other is_multiple fields should be sorted by sorting the values
        # for each book and using that as the sort key
        for field in ('tags', '#tags'):
            ae(cache.set_field(field, {1: ('b', 'a'), 2: ('c', 'y'), 3: ('b', 'z')}), {1, 2, 3})
            ae([1, 3, 2], cache.multisort([(field, True)], ids_to_sort=(1, 2, 3)))
            ae([2, 3, 1], cache.multisort([(field, False)], ids_to_sort=(1, 2, 3)))

        # Test tweak to sort dates by visible format
        from calibre.utils.config_base import Tweak

        ae(cache.set_field('pubdate', {1: p('2001-3-3'), 2: p('2002-2-3'), 3: p('2003-1-3')}), {1, 2, 3})
        ae([1, 2, 3], cache.multisort([('pubdate', True)]))
        with Tweak('gui_pubdate_display_format', 'MMM'), Tweak('sort_dates_using_visible_fields', True):
            c2 = self.init_cache()
            ae([3, 2, 1], c2.multisort([('pubdate', True)]))

        # Test bool sorting when not tristate
        cache.set_pref('bools_are_tristate', False)
        c2 = self.init_cache()
        ae([2, 3, 1], c2.multisort([('#yesno', True), ('id', False)]))

        # Test subsorting
        ae([3, 2, 1], cache.multisort([('identifiers', True), ('title', True)]), 'Subsort failed')
        from calibre.ebooks.metadata.book.base import Metadata

        for i in range(7):
            cache.create_book_entry(Metadata(f'title{i}'), apply_import_tags=False)
        cache.create_custom_column('one', 'CC1', 'int', False)
        cache.create_custom_column('two', 'CC2', 'int', False)
        cache.create_custom_column('three', 'CC3', 'int', False)
        cache.close()
        cache = self.init_cache()
        cache.set_field('#one', {(i + (5 * m)): m for m in (0, 1) for i in range(1, 6)})
        cache.set_field('#two', {i + (m * 3): m for m in (0, 1, 2) for i in (1, 2, 3)})
        cache.set_field('#two', {10: 2})
        cache.set_field('#three', {i: i for i in range(1, 11)})
        ae(
            list(range(1, 11)),
            cache.multisort([('#one', True), ('#two', True)], ids_to_sort=sorted(cache.all_book_ids())),
        )
        ae(
            [4, 5, 1, 2, 3, 7, 8, 9, 10, 6],
            cache.multisort([('#one', True), ('#two', False)], ids_to_sort=sorted(cache.all_book_ids())),
        )
        ae(
            [5, 4, 3, 2, 1, 10, 9, 8, 7, 6],
            cache.multisort([('#one', True), ('#two', False), ('#three', False)], ids_to_sort=sorted(cache.all_book_ids())),
        )

    # }}}

    def test_get_metadata(self):  # {{{
        "Test get_metadata() returns the same data for both backends"
        from calibre.library.database2 import LibraryDatabase2

        old = LibraryDatabase2(self.library_path)
        old_metadata = {i: old.get_metadata(i, index_is_id=True, get_cover=True, cover_as_data=True) for i in range(1, 4)}
        for mi in old_metadata.values():
            mi.format_metadata = dict(mi.format_metadata)
            if mi.formats:
                mi.formats = tuple(mi.formats)
        old.conn.close()
        old = None

        cache = self.init_cache(self.library_path)

        new_metadata = {i: cache.get_metadata(i, get_cover=True, cover_as_data=True) for i in range(1, 4)}
        cache = None
        for mi2, mi1 in zip(list(new_metadata.values()), list(old_metadata.values())):
            self.compare_metadata(mi1, mi2, exclude=('pages',))

    # }}}

    def test_serialize_metadata(self):  # {{{
        from calibre.library.field_metadata import fm_as_dict
        from calibre.utils.serialize import json_dumps, json_loads, msgpack_dumps, msgpack_loads

        cache = self.init_cache(self.library_path)
        fm = cache.field_metadata
        for d, l in ((json_dumps, json_loads), (msgpack_dumps, msgpack_loads)):
            fm2 = l(d(fm))
            self.assertEqual(fm_as_dict(fm), fm_as_dict(fm2))
        for i in range(1, 4):
            mi = cache.get_metadata(i, get_cover=True, cover_as_data=True)
            rmi = msgpack_loads(msgpack_dumps(mi))
            self.compare_metadata(mi, rmi, exclude='format_metadata has_cover formats id pages'.split())
            rmi = json_loads(json_dumps(mi))
            self.compare_metadata(mi, rmi, exclude='format_metadata has_cover formats id pages'.split())

    # }}}

    def test_get_cover(self):  # {{{
        "Test cover() returns the same data for both backends"
        from calibre.library.database2 import LibraryDatabase2

        old = LibraryDatabase2(self.library_path)
        covers = {i: old.cover(i, index_is_id=True) for i in old.all_ids()}
        old.conn.close()
        old = None
        cache = self.init_cache(self.library_path)
        for book_id, cdata in covers.items():
            self.assertEqual(cdata, cache.cover(book_id), 'Reading of cover failed')
            f = cache.cover(book_id, as_file=True)
            try:
                self.assertEqual(cdata, f.read() if f else f, 'Reading of cover as file failed')
            finally:
                if f:
                    f.close()
            if cdata:
                with open(cache.cover(book_id, as_path=True), 'rb') as f:
                    self.assertEqual(cdata, f.read(), 'Reading of cover as path failed')
            else:
                self.assertEqual(cdata, cache.cover(book_id, as_path=True), 'Reading of null cover as path failed')
        buf = BytesIO()
        self.assertFalse(cache.copy_cover_to(99999, buf), 'copy_cover_to() did not return False for non-existent book_id')
        self.assertFalse(cache.copy_cover_to(3, buf), 'copy_cover_to() did not return False for non-existent cover')

    # }}}

    def test_searching(self):  # {{{
        "Test searching returns the same data for both backends"
        from calibre.library.database2 import LibraryDatabase2

        old = LibraryDatabase2(self.library_path)
        oldvals = {
            query: set(old.search_getting_ids(query, ''))
            for query in (
                # Date tests
                'date:9/6/2011',
                'date:true',
                'date:false',
                'pubdate:1/9/2011',
                '#date:true',
                'date:<100_daysago',
                'date:>9/6/2011',
                '#date:>9/1/2011',
                '#date:=2011',
                # Number tests
                'rating:3',
                'rating:>2',
                'rating:=2',
                'rating:true',
                'rating:false',
                'rating:>4',
                'tags:#<2',
                'tags:#>7',
                'cover:false',
                'cover:true',
                '#float:>11',
                '#float:<1k',
                '#float:10.01',
                '#float:false',
                'series_index:1',
                'series_index:<3',
                # Bool tests
                '#yesno:true',
                '#yesno:false',
                '#yesno:_yes',
                '#yesno:_no',
                '#yesno:_empty',
                # Keypair tests
                'identifiers:true',
                'identifiers:false',
                'identifiers:test',
                'identifiers:test:false',
                'identifiers:test:one',
                'identifiers:t:n',
                'identifiers:=test:=two',
                'identifiers:x:y',
                'identifiers:z',
                # Text tests
                'title:="Title One"',
                'title:~title',
                '#enum:=one',
                '#enum:tw',
                '#enum:false',
                '#enum:true',
                'series:one',
                'tags:one',
                'tags:true',
                'tags:false',
                'uuid:2',
                'one',
                '20.02',
                '"publisher one"',
                '"my comments one"',
                'series_sort:one',
                # User categories
                '@Good Authors:One',
                '@Good Series.good tags:two',
                # Cover/Formats
                'cover:true',
                'cover:false',
                'formats:true',
                'formats:false',
                'formats:#>1',
                'formats:#=1',
                'formats:=fmt1',
                'formats:=fmt2',
                'formats:=fmt1 or formats:fmt2',
                '#formats:true',
                '#formats:false',
                '#formats:fmt1',
                '#formats:fmt2',
                '#formats:fmt1 and #formats:fmt2',
            )
        }
        old.conn.close()
        old = None

        cache = self.init_cache(self.cloned_library)
        for query, ans in oldvals.items():
            nr = cache.search(query, '')
            self.assertEqual(ans, nr, f'Old result: {ans!r} != New result: {nr!r} for search: {query}')

        # Test searching by id, which was introduced in the new backend
        self.assertEqual(cache.search('id:1', ''), {1})
        self.assertEqual(cache.search('id:>1', ''), {2, 3})

        # Numeric/rating searches with relops in the old db were incorrect, so
        # test them specifically here
        cache.set_field('rating', {1: 4, 2: 2, 3: 5})
        self.assertEqual(cache.search('rating:>2'), set())
        self.assertEqual(cache.search('rating:>=2'), {1, 3})
        self.assertEqual(cache.search('rating:<2'), {2})
        self.assertEqual(cache.search('rating:<=2'), {1, 2, 3})
        self.assertEqual(cache.search('rating:<=2'), {1, 2, 3})
        self.assertEqual(cache.search('rating:=2'), {1, 3})
        self.assertEqual(cache.search('rating:2'), {1, 3})
        self.assertEqual(cache.search('rating:!=2'), {2})

        cache.field_metadata.all_metadata()['#rating']['display']['allow_half_stars'] = True
        cache.set_field('#rating', {1: 3, 2: 4, 3: 9})
        self.assertEqual(cache.search('#rating:1'), set())
        self.assertEqual(cache.search('#rating:1.5'), {1})
        self.assertEqual(cache.search('#rating:>4'), {3})
        self.assertEqual(cache.search('#rating:2'), {2})

        # template searches
        # Test text search
        self.assertEqual(cache.search('template:"{#formats}#@#:t:fmt1"'), {1, 2})
        self.assertEqual(cache.search('template:"{authors}#@#:t:=Author One"'), {2})
        cache.set_field('pubdate', {1: p('2001-02-06'), 2: p('2001-10-06'), 3: p('2001-06-06')})
        cache.set_field('timestamp', {1: p('2002-02-06'), 2: p('2000-10-06'), 3: p('2001-06-06')})
        # Test numeric compare search
        self.assertEqual(
            cache.search("template:\"program: floor(days_between(field('pubdate'), field('timestamp')))#@#:n:>0\""),
            {2, 3},
        )
        # Test date search
        self.assertEqual(cache.search('template:{pubdate}#@#:d:<2001-09-01"'), {1, 3})
        # Test boolean search
        self.assertEqual(cache.search('template:{series}#@#:b:true'), {1, 2})
        self.assertEqual(cache.search('template:{series}#@#:b:false'), {3})

        # test primary search
        cache.set_field('title', {1: 'Gravity’s Raiñbow'})
        self.assertEqual(cache.search('title:"Gravity\'s Rainbow"'), {1})
        # Note that the old db searched uuid for un-prefixed searches, the new
        # db does not, for performance

    # }}}

    def test_get_categories(self):  # {{{
        "Check that get_categories() returns the same data for both backends"
        from calibre.library.database2 import LibraryDatabase2

        old = LibraryDatabase2(self.library_path)
        old_categories = old.get_categories()
        old.conn.close()
        cache = self.init_cache(self.library_path)
        new_categories = cache.get_categories()
        self.assertEqual(
            set(old_categories),
            set(new_categories),
            'The set of old categories is not the same as the set of new categories',
        )

        def compare_category(category, old, new):
            for attr in (
                'name',
                'original_name',
                'id',
                'count',
                'is_hierarchical',
                'is_editable',
                'is_searchable',
                'id_set',
                'avg_rating',
                'sort',
                'use_sort_as_name',
                'category',
            ):
                oval, nval = getattr(old, attr), getattr(new, attr)
                if (
                    (category in {'rating', '#rating'} and attr in {'id_set', 'sort'})
                    or (category == 'series' and attr == 'sort')  # Sorting is wrong in old
                    or (category == 'identifiers' and attr == 'id_set')
                    or (category == '@Good Series')  # Sorting is wrong in old
                    or (category == 'news' and attr in {'count', 'id_set'})
                    or (category == 'formats' and attr == 'id_set')
                ):
                    continue
                self.assertEqual(
                    oval,
                    nval,
                    f'The attribute {attr} for {old.name} in category {category} does not match. Old is {oval!r}, New is {nval!r}',
                )

        for category in old_categories:
            old, new = old_categories[category], new_categories[category]
            self.assertEqual(len(old), len(new), f'The number of items in the category {category} is not the same')
            for o, n in zip(old, new):
                compare_category(category, o, n)

    # }}}

    def test_get_formats(self):  # {{{
        "Test reading ebook formats using the format() method"
        from calibre.db.cache import NoSuchFormat
        from calibre.library.database2 import LibraryDatabase2

        old = LibraryDatabase2(self.library_path)
        ids = old.all_ids()
        lf = {i: set(old.formats(i, index_is_id=True).split(',')) if old.formats(i, index_is_id=True) else set() for i in ids}
        formats = {i: {f: old.format(i, f, index_is_id=True) for f in fmts} for i, fmts in lf.items()}
        old.conn.close()
        old = None
        cache = self.init_cache(self.library_path)
        for book_id, fmts in lf.items():
            self.assertEqual(fmts, set(cache.formats(book_id)), 'Set of formats is not the same')
            for fmt in fmts:
                old = formats[book_id][fmt]
                self.assertEqual(old, cache.format(book_id, fmt), 'Old and new format disagree')
                with cache.format(book_id, fmt, as_file=True) as f:
                    self.assertEqual(old, f.read(), 'Failed to read format as file')
                with open(cache.format(book_id, fmt, as_path=True, preserve_filename=True), 'rb') as f:
                    self.assertEqual(old, f.read(), 'Failed to read format as path')
                with open(cache.format(book_id, fmt, as_path=True), 'rb') as f:
                    self.assertEqual(old, f.read(), 'Failed to read format as path')

        buf = BytesIO()
        self.assertRaises(
            NoSuchFormat,
            cache.copy_format_to,
            99999,
            'X',
            buf,
            'copy_format_to() failed to raise an exception for non-existent book',
        )
        self.assertRaises(
            NoSuchFormat,
            cache.copy_format_to,
            1,
            'X',
            buf,
            'copy_format_to() failed to raise an exception for non-existent format',
        )
        fmt = cache.formats(1)[0]
        path = cache.format_abspath(1, fmt)
        changed_path = os.path.join(os.path.dirname(path), 'x' + os.path.basename(path))
        os.rename(path, changed_path)
        self.assertEqual(cache.format_abspath(1, fmt), path)
        self.assertFalse(os.path.exists(changed_path))

    # }}}

    def test_by_year(self):
        "Test grouping of books by date fields"
        cache = self.init_cache()
        self.assertEqual(cache.books_by_year(), {2011: {1, 2, 3}})
        self.assertEqual(cache.books_by_year(restrict_to_books={1, 2}), {2011: {1, 2}})
        self.assertEqual(cache.books_by_month(), {(2011, 8): {1}, (2011, 9): {2, 3}})
        self.assertEqual(cache.books_by_month(restrict_to_books={1, 2}), {(2011, 8): {1}, (2011, 9): {2}})
        self.assertEqual(cache.books_by_year(restrict_to_books=range(cache.backend.max_number_of_variables + 200)), {2011: {1, 2, 3}})
        cache.set_field('#date', {3: datetime.datetime(2001, 11, 10).astimezone(utc_tz)})
        self.assertEqual(cache.books_by_year('#date'), {2011: {1, 2}, 2001: {3}})
        self.assertEqual(cache.books_by_year('#date', restrict_to_books={1}), {2011: {1}})
        self.assertEqual(cache.books_by_month('#date'), {(2011, 9): {1, 2}, (2001, 11): {3}})
        self.assertEqual(cache.books_by_month('#date', restrict_to_books={1, 3}), {(2011, 9): {1}, (2001, 11): {3}})

    def test_author_sort_for_authors(self):  # {{{
        "Test getting the author sort for authors from the db"
        cache = self.init_cache()
        table = cache.fields['authors'].table
        table.set_sort_names({next(iter(table.id_map)): 'Fake Sort'}, cache.backend)

        authors = tuple(table.id_map.values())
        nval = cache.author_sort_from_authors(authors)
        self.assertIn('Fake Sort', nval)

        db = self.init_old()
        self.assertEqual(db.author_sort_from_authors(authors), nval)
        db.close()
        del db

    # }}}

    def test_get_next_series_num(self):  # {{{
        "Test getting the next series number for a series"
        cache = self.init_cache()
        cache.set_field('series', {3: 'test series'})
        cache.set_field('series_index', {3: 13})
        table = cache.fields['series'].table
        series = tuple(table.id_map.values())
        nvals = {s: cache.get_next_series_num_for(s) for s in series}
        db = self.init_old()
        self.assertEqual({s: db.get_next_series_num_for(s) for s in series}, nvals)
        db.close()

    # }}}

    def test_has_book(self):  # {{{
        "Test detecting duplicates"
        from calibre.ebooks.metadata.book.base import Metadata

        cache = self.init_cache()
        db = self.init_old()
        for title in cache.fields['title'].table.book_col_map.values():
            for x in (db, cache):
                self.assertTrue(x.has_book(Metadata(title)))
                self.assertTrue(x.has_book(Metadata(title.upper())))
                self.assertFalse(x.has_book(Metadata(title + 'XXX')))
                self.assertFalse(x.has_book(Metadata(title[:1])))
        db.close()

    # }}}

    def test_datetime(self):  # {{{
        "Test the reading of datetimes stored in the db"
        from calibre.db.tables import UNDEFINED_DATE, _c_speedup, c_parse
        from calibre.utils.date import parse_date

        # First test parsing of string to UTC time
        for raw in (
            '2013-07-22 15:18:29+05:30',
            '  2013-07-22 15:18:29+00:00',
            '2013-07-22 15:18:29',
            '2003-09-21 23:30:00-06:00',
        ):
            self.assertTrue(_c_speedup(raw))
            ctime = c_parse(raw)
            pytime = parse_date(raw, assume_utc=True)
            self.assertEqual(ctime, pytime)

        self.assertEqual(c_parse(2003).year, 2003)
        for x in (None, '', 'abc'):
            self.assertEqual(UNDEFINED_DATE, c_parse(x))

    # }}}

    def test_restrictions(self):  # {{{
        "Test searching with and without restrictions"
        cache = self.init_cache()
        se = self.assertSetEqual
        se(cache.all_book_ids(), cache.search(''))
        se({1, 2}, cache.search('', 'not authors:=Unknown'))
        se(set(), cache.search('authors:=Unknown', 'not authors:=Unknown'))
        se({2}, cache.search('not authors:"=Author Two"', 'not authors:=Unknown'))
        se({2}, cache.search('not authors:"=Author Two"', book_ids={1, 2}))
        se({2}, cache.search('not authors:"=Author Two"', 'not authors:=Unknown', book_ids={1, 2, 3}))
        se(set(), cache.search('authors:=Unknown', 'not authors:=Unknown', book_ids={1, 2, 3}))
        se(cache.all_book_ids(), cache.books_in_virtual_library(''))
        se(cache.all_book_ids(), cache.books_in_virtual_library('does not exist'))
        cache.set_pref('virtual_libraries', {'1': 'title:"=Title One"', '12': 'id:1 or id:2'})
        se({2}, cache.books_in_virtual_library('1'))
        se({1, 2}, cache.books_in_virtual_library('12'))
        se({1}, cache.books_in_virtual_library('12', 'id:1'))
        se({2}, cache.books_in_virtual_library('1', 'id:1 or id:2'))

    # }}}

    def test_search_caching(self):  # {{{
        "Test caching of searches"
        from calibre.db.search import LRUCache

        class TestCache(LRUCache):
            hit_counter = 0
            miss_counter = 0

            def get(self, key, default=None):
                ans = LRUCache.get(self, key, default=default)
                if ans is not None:
                    self.hit_counter += 1
                else:
                    self.miss_counter += 1

            @property
            def cc(self):
                self.hit_counter = self.miss_counter = 0

            @property
            def counts(self):
                return self.hit_counter, self.miss_counter

        cache = self.init_cache()
        cache._search_api.cache = c = TestCache()

        ae = self.assertEqual

        def test(hit, result, *args, **kw):
            c.cc
            num = kw.get('num', 2)
            ae(cache.search(*args), result)
            ae(c.counts, (num, 0) if hit else (0, num))
            c.cc

        test(False, {3}, 'Unknown')
        test(True, {3}, 'Unknown')
        test(True, {3}, 'Unknown')
        cache._search_api.MAX_CACHE_UPDATE = 0
        cache.set_field('title', {3: 'xxx'})
        test(False, {3}, 'Unknown')  # cache cleared
        test(True, {3}, 'Unknown')
        c.limit = 5
        for i in range(6):
            test(False, set(), f'nomatch_{i}')
        test(False, {3}, 'Unknown')  # cached search expired
        test(False, {3}, '', 'unknown', num=1)
        test(True, {3}, '', 'unknown', num=1)
        test(True, {3}, 'Unknown', 'unknown', num=1)
        cache._search_api.MAX_CACHE_UPDATE = 100
        test(False, {2, 3}, 'title:=xxx or title:"=Title One"')
        cache.set_field('publisher', {3: 'ppppp', 2: 'other'})
        # Test cache update worked
        test(True, {2, 3}, 'title:=xxx or title:"=Title One"')

    # }}}

    def test_proxy_metadata(self):  # {{{
        "Test the ProxyMetadata object used for composite columns"
        from calibre.ebooks.metadata.book.base import STANDARD_METADATA_FIELDS

        cache = self.init_cache()
        for book_id in cache.all_book_ids():
            mi = cache.get_metadata(book_id, get_user_categories=True)
            pmi = cache.get_proxy_metadata(book_id)
            self.assertSetEqual(set(mi.custom_field_keys()), set(pmi.custom_field_keys()))

            for field in STANDARD_METADATA_FIELDS | {'#series_index'}:
                if field == 'formats':

                    def f(x):
                        return x if x is None else tuple(x)

                else:

                    def f(x):
                        return x

                self.assertEqual(
                    f(getattr(mi, field)),
                    f(getattr(pmi, field)),
                    f'Standard field: {field} not the same for book {book_id}',
                )
                self.assertEqual(
                    mi.format_field(field),
                    pmi.format_field(field),
                    f'Standard field format: {field} not the same for book {book_id}',
                )

                def f(x):
                    try:
                        x.pop('rec_index', None)
                    except Exception:
                        pass
                    return x

                if field != '#series_index':
                    v = pmi.get_standard_metadata(field)
                    self.assertTrue(v is None or isinstance(v, dict))
                    self.assertEqual(
                        f(mi.get_standard_metadata(field, False)),
                        f(v),
                        f'get_standard_metadata() failed for field {field}',
                    )
            for field, meta in cache.field_metadata.custom_iteritems():
                if meta['datatype'] != 'composite':
                    self.assertEqual(
                        f(getattr(mi, field)),
                        f(getattr(pmi, field)),
                        f'Custom field: {field} not the same for book {book_id}',
                    )
                    self.assertEqual(
                        mi.format_field(field),
                        pmi.format_field(field),
                        f'Custom field format: {field} not the same for book {book_id}',
                    )

        # Test handling of recursive templates
        cache.create_custom_column('comp2', 'comp2', 'composite', False, display={'composite_template': '{title}'})
        cache.create_custom_column('comp1', 'comp1', 'composite', False, display={'composite_template': 'foo{#comp2}'})
        cache.close()
        cache = self.init_cache()
        mi, pmi = cache.get_metadata(1), cache.get_proxy_metadata(1)
        self.assertEqual(mi.get('#comp1'), pmi.get('#comp1'))

        # Test overridden Metadata methods

        self.assertTrue(pmi.has_key('tags') == mi.has_key('tags'))

        self.assertFalse(pmi.has_key('taggs'), 'taggs attribute')
        self.assertTrue(pmi.has_key('taggs') == mi.has_key('taggs'))

        self.assertSetEqual(set(pmi.custom_field_keys()), set(mi.custom_field_keys()))

        self.assertEqual(pmi.get_extra('#series', 0), 3)
        self.assertEqual(pmi.get_extra('#series', 0), mi.get_extra('#series', 0))

        self.assertDictEqual(pmi.get_identifiers(), {'test': 'two'})
        self.assertDictEqual(pmi.get_identifiers(), mi.get_identifiers())

        self.assertTrue(pmi.has_identifier('test'))
        self.assertTrue(pmi.has_identifier('test') == mi.has_identifier('test'))

        self.assertListEqual(list(pmi.custom_field_keys()), list(mi.custom_field_keys()))

        # ProxyMetadata has the virtual fields while Metadata does not.
        self.assertSetEqual(
            set(pmi.all_field_keys()) - {'id', 'series_sort', 'path', 'in_tag_browser', 'sort', 'ondevice', 'au_map', 'marked', '#series_index'},
            set(mi.all_field_keys()),
        )

        # mi.get_standard_metadata() doesn't include the rec_index metadata key
        fm_pmi = pmi.get_standard_metadata('series')
        fm_pmi.pop('rec_index')
        self.assertDictEqual(fm_pmi, mi.get_standard_metadata('series', make_copy=False))

        # The ProxyMetadata versions don't include the values. Note that the mi
        # version of get_standard_metadata won't return custom columns while the
        # ProxyMetadata version will
        fm_mi = mi.get_user_metadata('#series', make_copy=False)
        fm_mi.pop('#extra#')
        fm_mi.pop('#value#')
        self.assertDictEqual(pmi.get_standard_metadata('#series'), fm_mi)
        self.assertDictEqual(pmi.get_user_metadata('#series'), fm_mi)

        fm_mi = mi.get_all_user_metadata(make_copy=False)
        for one in fm_mi:
            fm_mi[one].pop('#extra#', None)
            fm_mi[one].pop('#value#', None)
        self.assertDictEqual(pmi.get_all_user_metadata(make_copy=False), fm_mi)

        # Check the unimplemented methods
        self.assertRaises(NotImplementedError, lambda: 'foo' in pmi)
        self.assertRaises(NotImplementedError, pmi.set, 'a', 'a')
        self.assertRaises(NotImplementedError, pmi.set_identifiers, 'a')
        self.assertRaises(NotImplementedError, pmi.set_identifier, 'a', 'a')
        self.assertRaises(NotImplementedError, pmi.all_non_none_fields)
        self.assertRaises(NotImplementedError, pmi.set_all_user_metadata, {})
        self.assertRaises(NotImplementedError, pmi.set_user_metadata, 'a', {})
        self.assertRaises(NotImplementedError, pmi.remove_stale_user_metadata, {})
        self.assertRaises(NotImplementedError, pmi.template_to_attribute, {}, {})
        self.assertRaises(NotImplementedError, pmi.smart_update, {})

    # }}}

    def test_marked_field(self):  # {{{
        "Test the marked field"
        db = self.init_legacy()
        db.set_marked_ids({3: 1, 2: 3})
        ids = [1, 2, 3]
        db.multisort([('marked', True)], only_ids=ids)
        self.assertListEqual([1, 3, 2], ids)
        db.multisort([('marked', False)], only_ids=ids)
        self.assertListEqual([2, 3, 1], ids)

    # }}}

    def test_composites(self):  # {{{
        "Test sorting and searching in composite columns"
        cache = self.init_cache()
        cache.create_custom_column('mult', 'CC1', 'composite', True, display={'composite_template': 'b,a,c'})
        cache.create_custom_column('single', 'CC2', 'composite', False, display={'composite_template': 'b,a,c'})
        cache.create_custom_column('number', 'CC3', 'composite', False, display={'composite_template': '{#float}', 'composite_sort': 'number'})
        cache.create_custom_column(
            'size',
            'CC4',
            'composite',
            False,
            display={'composite_template': '{#float:human_readable()}', 'composite_sort': 'number'},
        )
        cache.create_custom_column(
            'ccdate',
            'CC5',
            'composite',
            False,
            display={
                'composite_template': "{:'format_date(raw_field('pubdate'), 'dd-MM-yy')'}",
                'composite_sort': 'date',
            },
        )
        cache.create_custom_column('bool', 'CC6', 'composite', False, display={'composite_template': '{#yesno}', 'composite_sort': 'bool'})
        cache.create_custom_column('ccm', 'CC7', 'composite', True, display={'composite_template': '{#tags}'})
        cache.create_custom_column('ccp', 'CC8', 'composite', True, display={'composite_template': '{publisher}'})
        cache.create_custom_column('ccf', 'CC9', 'composite', True, display={'composite_template': "{:'approximate_formats()'}"})

        cache = self.init_cache()
        # Test searching
        self.assertEqual({1, 2, 3}, cache.search('#mult:=a'))
        self.assertEqual(set(), cache.search('#mult:=b,a,c'))
        self.assertEqual({1, 2, 3}, cache.search('#single:=b,a,c'))
        self.assertEqual(set(), cache.search('#single:=b'))

        # Test numeric sorting
        cache.set_field('#float', {1: 2, 2: 10, 3: 0.0001})
        self.assertEqual([3, 1, 2], cache.multisort([('#number', True)]))
        cache.set_field('#float', {1: 3, 2: 2 * 1024, 3: 1 * 1024 * 1024})
        self.assertEqual([1, 2, 3], cache.multisort([('#size', True)]))

        # Test date sorting
        cache.set_field('pubdate', {1: p('2001-02-06'), 2: p('2001-10-06'), 3: p('2001-06-06')})
        self.assertEqual([1, 3, 2], cache.multisort([('#ccdate', True)]))

        # Test bool sorting
        self.assertEqual([2, 1, 3], cache.multisort([('#bool', True)]))

        # Test is_multiple sorting
        cache.set_field('#tags', {1: 'b, a, c', 2: 'a, b, c', 3: 'a, c, b'})
        self.assertEqual([1, 2, 3], cache.multisort([('#ccm', True)]))

        # Test that lock downgrading during update of search cache works
        self.assertEqual(cache.search('#ccp:One'), {2})
        cache.set_field('publisher', {2: 'One', 1: 'One'})
        self.assertEqual(cache.search('#ccp:One'), {1, 2})

        self.assertEqual(cache.search('#ccf:FMT1'), {1, 2})
        cache.remove_formats({1: ('FMT1',)})
        self.assertEqual('FMT2', cache.field_for('#ccf', 1))

    # }}}

    def test_find_identical_books(self):  # {{{
        "Test find_identical_books"
        from calibre.db.utils import find_identical_books
        from calibre.ebooks.metadata.book.base import Metadata

        # 'find_identical_books': [(,), (Metadata('unknown'),), (Metadata('xxxx'),)],
        cache = self.init_cache(self.library_path)
        cache.set_field('languages', {1: ('fra', 'deu')})
        data = cache.data_for_find_identical_books()
        lm = cache.get_metadata(1)
        lm2 = cache.get_metadata(1)
        lm2.languages = ['eng']
        for mi, books in (
            (Metadata('title one', ['author one']), {2}),
            (Metadata('Unknown', ['Unknown']), {3}),
            (Metadata('title two', ['author one']), {1}),
            (lm, {1}),
            (lm2, set()),
        ):
            self.assertEqual(books, cache.find_identical_books(mi))
            self.assertEqual(books, find_identical_books(mi, data))

    # }}}

    def test_last_read_positions(self):  # {{{
        cache = self.init_cache(self.library_path)
        self.assertFalse(cache.get_last_read_positions(1, 'x', 'u'))
        self.assertRaises(Exception, cache.set_last_read_position, 12, 'x', cfi='c')
        epoch = time()
        cache.set_last_read_position(1, 'EPUB', 'user', 'device', 'cFi', epoch, 0.3)
        self.assertFalse(cache.get_last_read_positions(1, 'x', 'u'))
        self.assertEqual(
            cache.get_last_read_positions(1, 'ePuB', 'user'),
            ({'epoch': epoch, 'device': 'device', 'cfi': 'cFi', 'pos_frac': 0.3, 'format': 'EPUB', 'user': 'user'},),
        )
        cache.set_last_read_position(1, 'EPUB', 'user', 'device')
        self.assertFalse(cache.get_last_read_positions(1, 'ePuB', 'user'))

    # }}}

    def test_storing_conversion_options(self):  # {{{
        cache = self.init_cache(self.library_path)
        opts = {1: b'binary', 2: 'unicode'}
        cache.set_conversion_options(opts, 'PIPE')
        for book_id, val in opts.items():
            got = cache.conversion_options(book_id, 'PIPE')
            if not isinstance(val, bytes):
                val = val.encode('utf-8')
            self.assertEqual(got, val)

    # }}}

    def test_template_db_functions(self):  # {{{
        from calibre.ebooks.metadata.book.formatter import SafeFormat

        formatter = SafeFormat()

        db = self.init_cache(self.library_path)
        db.create_custom_column('mult', 'CC1', 'composite', True, display={'composite_template': 'b,a,c'})
        db.create_custom_column('pages', 'Pages', 'int', False)
        db.set_pages(2, 100, format='FMT1')
        db.set_pages(2, 100, format='FMT2')
        db.set_last_read_position(2, 'FMT1', pos_frac=0.25, cfi='epubcfi(/2)', epoch=2)
        db.set_last_read_position(2, 'FMT2', pos_frac=0.35, cfi='epubcfi(/2)', epoch=1)
        db.close()
        db = self.init_cache(self.library_path)
        db.set_field('#pages', {1: '11', 2: '22', 3: '33'})
        mi = db.get_metadata(2)

        # test reading_progress
        def trp(expected, args=''):
            self.assertEqual(expected, formatter.safe_format(f'{{id:reading_progress({args})}}', {}, 'TEMPLATE ERROR', mi))

        trp('25 / 100')
        trp('25%', ',percent')
        trp('25%', '_,percent')
        trp('0.35', ',pos_frac,furthest')
        trp('0.25', ',pos_frac,furthest,fmt1')
        # test width_from_pages
        v = formatter.safe_format('{#pages:width_from_pages(2,2,0.5)}', {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(v, '1.0')
        v = formatter.safe_format('{#pages:width_from_pages()}', {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(int(float(v) * 1000), 26)
        v = formatter.safe_format('{#pages:width_from_pages(1500)}', {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(int(float(v) * 1000), 26)
        v = formatter.safe_format('{#pages:width_from_pages(1500,2)}', {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(int(float(v) * 1000), 26)
        v = formatter.safe_format('{#pages:width_from_pages(1500,2,11)}', {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(int(float(v) * 1000), 26)

        # test counting books matching the search
        v = formatter.safe_format('program: book_count("series:true", 0)', {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(v, '2')

        # test python templates
        v = formatter.safe_format('python:\ndef evaluate(book, ctx): return "x"', {}, 'TEMPLATE ERROR', mi)
        if os.environ.get('CALIBRE_ALLOW_PYTHON_TEMPLATES', '1') == '1':
            self.assertEqual(v, 'x')
        else:
            self.assertTrue(v.startswith('TEMPLATE ERROR'))

        # test counting books when none match the search
        v = formatter.safe_format('program: book_count("series:afafaf", 0)', {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(v, '0')

        # test is_multiple values
        v = formatter.safe_format('program: book_values("tags", "tags:true", ",", 0)', {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(set(v.split(',')), {'Tag One', 'News', 'Tag Two'})

        # test not is_multiple values
        v = formatter.safe_format('program: book_values("series", "series:true", ",", 0)', {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(v, 'A Series One')

        # test returning values for a column not searched for
        v = formatter.safe_format('program: book_values("tags", "series:\\"A Series One\\"", ",", 0)', {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(set(v.split(',')), {'Tag One', 'News', 'Tag Two'})

        # test getting a singleton value from books where the column is empty
        v = formatter.safe_format('program: book_values("series", "series:false", ",", 0)', {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(v, '')

        # test getting a multiple value from books where the column is empty
        v = formatter.safe_format('program: book_values("tags", "tags:false", ",", 0)', {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(v, '')

        # test fetching an unknown column
        v = formatter.safe_format('program: book_values("taaags", "tags:false", ",", 0)', {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(v, "TEMPLATE ERROR The column taaags doesn't exist")

        # test finding all books
        v = formatter.safe_format('program: book_values("id", "title:true", ",", 0)', {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(set(v.split(',')), {'1', '2', '3'})

        # test getting value of a composite
        v = formatter.safe_format('program: book_values("#mult", "id:1", ",", 0)', {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(set(v.split(',')), {'b', 'c', 'a'})

        # test getting value of a custom float
        v = formatter.safe_format('program: book_values("#float", "title:true", ",", 0)', {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(set(v.split(',')), {'20.02', '10.01'})

        # test getting value of an int (rating)
        v = formatter.safe_format('program: book_values("rating", "title:true", ",", 0)', {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(set(v.split(',')), {'4', '6'})

    # }}}

    @unittest.skipIf(os.environ.get('CALIBRE_ALLOW_PYTHON_TEMPLATES', '1') != '1', 'Python templates disallowed')
    def test_python_templates(self):  # {{{
        from calibre.ebooks.metadata.book.formatter import SafeFormat

        formatter = SafeFormat()

        # need an empty metadata object to pass to the formatter
        db = self.init_legacy(self.library_path)
        mi = db.get_metadata(1)

        # test counting books matching a search
        template = '''python:
def evaluate(book, ctx):
    ids = ctx.db.new_api.search("series:true")
    return str(len(ids))
'''
        v = formatter.safe_format(template, {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(v, '2')

        # test counting books when none match the search
        template = '''python:
def evaluate(book, ctx):
    ids = ctx.db.new_api.search("series:afafaf")
    return str(len(ids))
'''
        v = formatter.safe_format(template, {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(v, '0')

        # test is_multiple values
        template = '''python:
def evaluate(book, ctx):
    tags = ctx.db.new_api.all_field_names('tags')
    return ','.join(list(tags))
'''
        v = formatter.safe_format(template, {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(set(v.split(',')), {'Tag One', 'News', 'Tag Two'})

        # test using a custom context class
        template = '''python:
def evaluate(book, ctx):
    tags = ctx.db.new_api.all_field_names('tags')
    return ','.join(list(ctx.helper_function(tags)))
'''
        from calibre.utils.formatter import PythonTemplateContext

        class CustomContext(PythonTemplateContext):
            def helper_function(self, arg):
                s = set(arg)
                s.add('helper called')
                return s

        v = formatter.safe_format(template, {}, 'TEMPLATE ERROR', mi, python_context_object=CustomContext())
        self.assertEqual(set(v.split(',')), {'Tag One', 'News', 'Tag Two', 'helper called'})

        # test is_multiple values
        template = '''python:
def evaluate(book, ctx):
    tags = ctx.db.new_api.all_field_names('tags')
    return ','.join(list(tags))
'''
        v = formatter.safe_format(template, {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(set(v.split(',')), {'Tag One', 'News', 'Tag Two'})

        # test calling a python stored template from a GPM template
        from calibre.utils.formatter_functions import load_user_template_functions, unload_user_template_functions

        load_user_template_functions(
            'aaaaa',
            [
                [
                    'python_stored_template',
                    '',
                    0,
                    '''python:
def evaluate(book, ctx):
    tags = set(ctx.db.new_api.all_field_names('tags'))
    tags.add(ctx.arguments[0])
    return ','.join(list(tags))
''',
                ]
            ],
            None,
        )
        v = formatter.safe_format('program: python_stored_template("one argument")', {}, 'TEMPLATE ERROR', mi)
        unload_user_template_functions('aaaaa')
        self.assertEqual(set(v.split(',')), {'Tag One', 'News', 'Tag Two', 'one argument'})

        # test that recursive template() calls inherit the python template policy
        # The outer GPM calls template() with a python template string argument.
        # The single-quoted string spans a newline; re.DOTALL allows this.
        recursive_python = "program: template('python:\ndef evaluate(book, ctx): return \"recursive_ran\"')"

        # with allow_python_templates=True, the recursive template should succeed
        formatter2 = SafeFormat()
        formatter2.allow_python_templates = True
        v = formatter2.safe_format(recursive_python, {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(v, 'recursive_ran')

        # with allow_python_templates=False, the recursive template should be blocked.
        # The inner safe_format (called inside template()) returns the disallow error
        # prefixed with its own error_value 'TEMPLATE'; the outer formatter returns
        # that string as a normal result rather than its own 'TEMPLATE ERROR' prefix.
        formatter3 = SafeFormat()
        formatter3.allow_python_templates = False
        v = formatter3.safe_format(recursive_python, {}, 'TEMPLATE ERROR', mi)
        self.assertNotEqual(v, 'recursive_ran')
        self.assertTrue(v.startswith('TEMPLATE'))

        # test that eval() inherits the python template policy.
        # eval() uses EvalFormatter with book=None, so the python template must
        # not access ctx.db.
        eval_python = "program: eval('python:\ndef evaluate(book, ctx): return \"eval_ran\"')"

        formatter4 = SafeFormat()
        formatter4.allow_python_templates = True
        v = formatter4.safe_format(eval_python, {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(v, 'eval_ran')

        formatter5 = SafeFormat()
        formatter5.allow_python_templates = False
        v = formatter5.safe_format(eval_python, {}, 'TEMPLATE ERROR', mi)
        self.assertNotEqual(v, 'eval_ran')
        self.assertTrue(v.startswith('EVAL'))

        # test that re_group() inherits the python template policy.
        # The template substitution arg to re_group() is evaluated by EvalFormatter
        # with book=None, so the python template must not access ctx.db.
        re_group_python = "program: re_group('hello', '(hello)', 'python:\ndef evaluate(book, ctx): return \"re_group_ran\"')"

        formatter6 = SafeFormat()
        formatter6.allow_python_templates = True
        v = formatter6.safe_format(re_group_python, {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(v, 're_group_ran')

        formatter7 = SafeFormat()
        formatter7.allow_python_templates = False
        v = formatter7.safe_format(re_group_python, {}, 'TEMPLATE ERROR', mi)
        self.assertNotEqual(v, 're_group_ran')

        # test that list_re_group() inherits the python template policy.
        list_re_group_python = "program: list_re_group('hello', ',', 'hello', '(hello)', 'python:\ndef evaluate(book, ctx): return \"list_re_group_ran\"')"

        formatter8 = SafeFormat()
        formatter8.allow_python_templates = True
        v = formatter8.safe_format(list_re_group_python, {}, 'TEMPLATE ERROR', mi)
        self.assertEqual(v, 'list_re_group_ran')

        formatter9 = SafeFormat()
        formatter9.allow_python_templates = False
        v = formatter9.safe_format(list_re_group_python, {}, 'TEMPLATE ERROR', mi)
        self.assertNotEqual(v, 'list_re_group_ran')

    # }}}

    def test_cover_cache(self):
        from calibre.gui2.library.caches import test_cover_cache

        test_cover_cache(self)



# License: GPLv3 Copyright: 2009, John Schember <john@nachtimwald.com>

"""
Interface defining the necessary public functions for a pdb format writer.
"""


class FormatWriter:
    def __init__(self, opts, log):
        raise NotImplementedError()

    def write_content(self, oeb_book, output_stream, metadata=None):
        raise NotImplementedError()



#!/usr/bin/env python
# License: GPLv3 Copyright: 2010, Kovid Goyal <kovid@kovidgoyal.net>

import numbers
import os
from functools import partial

from calibre.gui2 import Dispatcher, choose_dir, error_dialog
from calibre.gui2.actions import InterfaceAction
from calibre.utils.config import prefs
from calibre.utils.localization import _


class SaveToDiskAction(InterfaceAction):
    name = 'Save To Disk'
    action_spec = (_('Save to disk'), 'save.png', _('Export e-book files from the calibre library'), _('S'))
    action_type = 'current'
    action_add_menu = True
    action_menu_clone_qaction = True

    def genesis(self):
        self.qaction.triggered.connect(self.save_to_disk)
        self.save_menu = self.qaction.menu()
        assert self.save_menu is not None
        cm = partial(self.create_menu_action, self.save_menu)
        cm('single dir', _('Save to disk in a single folder'), triggered=partial(self.save_to_single_dir, False))
        cm(
            'single format',
            _('Save only %s format to disk') % prefs['output_format'].upper(),
            triggered=partial(self.save_single_format_to_disk, False),
        )
        cm(
            'single dir and format',
            _('Save only %s format to disk in a single folder') % prefs['output_format'].upper(),
            triggered=partial(self.save_single_fmt_to_single_dir, False),
        )
        cm('specific format', _('Save single format to disk...'), triggered=self.save_specific_format_disk)

    def location_selected(self, loc):
        enabled = loc == 'library'
        assert self.save_menu is not None
        for action in list(self.save_menu.actions())[1:]:
            action.setEnabled(enabled)

    def reread_prefs(self):
        assert self.save_menu is not None
        self.save_menu.actions()[2].setText(_('Save only %s format to disk') % prefs['output_format'].upper())
        self.save_menu.actions()[3].setText(_('Save only %s format to disk in a single folder') % prefs['output_format'].upper())

    def save_single_format_to_disk(self, checked):
        self.save_to_disk(checked, False, prefs['output_format'])

    def save_specific_format_disk(self):
        rb = self.gui.iactions['Remove Books']
        ids = rb._get_selected_ids(err_title=_('Cannot save to disk'))
        if not ids:
            return
        fmts = rb._get_selected_formats(_('Choose format to save to disk'), ids, single=True, add_cover=True)
        if not fmts:
            return
        fmt = list(fmts)[0]
        self.save_to_disk(False, False, fmt)

    def save_to_single_dir(self, checked):
        self.save_to_disk(checked, True)

    def save_single_fmt_to_single_dir(self, *args):
        self.save_to_disk(False, single_dir=True, single_format=prefs['output_format'])

    def save_to_disk(self, checked, single_dir=False, single_format=None, rows=None, write_opf=None, save_cover=None, saver_class=None, path=None):
        if rows is None:
            rows = self.gui.current_view().selectionModel().selectedRows()
        if not rows or len(rows) == 0:
            return error_dialog(self.gui, _('Cannot save to disk'), _('No books selected'), show=True)
        if not path:
            path = choose_dir(self.gui, 'save to disk dialog', _('Choose destination folder'))
            if not path:
                return
        dpath = os.path.abspath(path).replace('/', os.sep) + os.sep
        lpath = self.gui.library_view.model().db.library_path.replace('/', os.sep) + os.sep
        if dpath.startswith(lpath):
            return error_dialog(
                self.gui,
                _('Not allowed'),
                _(
                    'You are trying to save files into the calibre '
                    'library. This can cause corruption of your '
                    'library. Save to disk is meant to export '
                    'files from your calibre library elsewhere.'
                ),
                show=True,
            )

        if self.gui.current_view() is self.gui.library_view:
            from calibre.gui2.save import Saver
            from calibre.library.save_to_disk import config

            opts = config().parse()
            if single_format is not None:
                opts.formats = single_format
                # Special case for Kindle annotation files
                if single_format.lower() in ['mbp', 'pdr', 'tan']:
                    opts.to_lowercase = False
                    opts.save_cover = False
                    opts.write_opf = False
                    opts.save_extra_files = False
                    opts.template = opts.send_template
                elif single_format == '..cover..':
                    opts.save_cover = True
            opts.single_dir = single_dir
            if write_opf is not None:
                opts.write_opf = write_opf
            if save_cover is not None:
                opts.save_cover = save_cover
            book_ids = set(map(self.gui.library_view.model().id, rows))
            if saver_class is None:
                saver_class = Saver
            assert issubclass(saver_class, Saver)
            saver_class(book_ids, self.gui.current_db, opts, path, parent=self.gui, pool=self.gui.spare_pool())
        else:
            paths = self.gui.current_view().model().paths(rows)
            self.gui.device_manager.save_books(Dispatcher(self.books_saved), paths, path)

    def save_library_format_by_ids(self, book_ids, fmt, single_dir=True, path=None, saver_class=None):
        if isinstance(book_ids, numbers.Integral):
            book_ids = [book_ids]
        rows = list(self.gui.library_view.ids_to_rows(book_ids).values())
        rows = [self.gui.library_view.model().index(r, 0) for r in rows]
        self.save_to_disk(True, single_dir=single_dir, single_format=fmt, rows=rows, write_opf=False, save_cover=False, path=path, saver_class=saver_class)

    def books_saved(self, job):
        if job.failed:
            return self.gui.device_job_exception(job)



# strip out a set of nuisance html attributes that can mess up rendering in RSS feeds
import re

from lxml_html_clean import Cleaner

bad_attrs = ['width', 'height', 'style', '[-a-z]*color', 'background[-a-z]*', 'on*']
single_quoted = "'[^']+'"
double_quoted = '"[^"]+"'
non_space = '[^ "\'>]+'
htmlstrip = re.compile(
    '<'  # open
    '([^>]+) '  # prefix
    '(?:{}) *'.format('|'.join(bad_attrs))  # undesirable attributes
    + f'= *(?:{non_space}|{single_quoted}|{double_quoted})'  # value
    + '([^>]*)'  # postfix
    '>',  # end
    re.I,
)


def clean_attributes(html):
    while htmlstrip.search(html):
        html = htmlstrip.sub(r'<\1\2>', html)
    return html


def normalize_spaces(s):
    if not s:
        return ''
    '''replace any sequence of whitespace
    characters with a single space'''
    return ' '.join(s.split())


html_cleaner = Cleaner(
    scripts=True,
    javascript=True,
    comments=True,
    style=True,
    links=True,
    meta=False,
    add_nofollow=False,
    page_structure=False,
    processing_instructions=True,
    embedded=False,
    frames=False,
    forms=False,
    annoying_tags=False,
    remove_tags=(),
    remove_unknown_tags=False,
    safe_attrs_only=False,
)



#!/usr/bin/env python
# License: GPLv3 Copyright: 2010, Kovid Goyal <kovid@kovidgoyal.net>

import functools
import numbers
import os
import re
import sys
import time
import traceback
from collections import defaultdict, namedtuple
from collections.abc import Callable
from itertools import groupby
from typing import cast

from qt.core import (
    QAbstractTableModel,
    QApplication,
    QColor,
    QDateTime,
    QFont,
    QFontMetrics,
    QIcon,
    QImage,
    QLocale,
    QModelIndex,
    QPainter,
    QPixmap,
    Qt,
    QTimer,
    pyqtSignal,
)

from calibre import fit_image, human_readable, prepare_string_for_xml, strftime
from calibre.constants import DEBUG, config_dir, dark_link_color, filesystem_encoding
from calibre.db.search import CONTAINS_MATCH, EQUALS_MATCH, REGEXP_MATCH, _match
from calibre.db.utils import force_to_bool
from calibre.devices.interface import BookList
from calibre.ebooks.metadata import authors_to_string, fmt_sidx, string_to_authors, title_sort
from calibre.ebooks.metadata.book.formatter import SafeFormat
from calibre.gui2 import error_dialog, is_dark_theme, qapplication_or_fail, simple_excepthook
from calibre.gui2.library import DEFAULT_SORT
from calibre.library.coloring import color_row_key
from calibre.library.save_to_disk import find_plugboard
from calibre.ptempfile import PersistentTemporaryFile
from calibre.utils.config import device_prefs, prefs, tweaks
from calibre.utils.date import UNDEFINED_DATE, dt_factory, is_date_undefined, qt_from_dt, qt_to_dt
from calibre.utils.icu import sort_key
from calibre.utils.localization import _, calibre_langcode_to_name, ngettext
from calibre.utils.resources import get_path as P
from calibre.utils.search_query_parser import ParseException, SearchQueryParser
from calibre_extensions.imageops import load_from_data_without_gil

Counts = namedtuple('Counts', 'library_total total current')

TIME_FMT = '%d %b %Y'

ALIGNMENT_MAP = {
    'left': Qt.AlignmentFlag.AlignLeft,
    'right': Qt.AlignmentFlag.AlignRight,
    'center': Qt.AlignmentFlag.AlignHCenter,
}


def render_pin(color=None, save_to=None):
    app = qapplication_or_fail()
    if color is None:
        color = '#00b000' if app.is_dark_theme else 'green'
    svg = P('pin-template.svg', data=True).replace(b'fill:#f39509', ('fill:' + color).encode('utf-8'))
    pm = QPixmap()
    dpr = app.devicePixelRatio()
    pm.setDevicePixelRatio(dpr)
    pm.loadFromData(svg, 'svg')
    if save_to:
        pm.save(save_to)
    return pm


def group_numbers(numbers):
    for k, g in groupby(enumerate(sorted(numbers)), lambda i_x: i_x[0] - i_x[1]):
        first = None
        for last in g:
            if first is None:
                first = last[1]
        yield first, last[1]


class ColumnColor:  # {{{
    def __init__(self, formatter):
        self.mi = None
        self.formatter = formatter

    def __call__(self, id_, key, fmt, db, color_cache, template_cache):
        key += str(hash(fmt))
        if id_ in color_cache and key in color_cache[id_]:
            self.mi = None
            color = color_cache[id_][key]
            if color.isValid():
                return color
            return None
        try:
            if self.mi is None:
                self.mi = db.new_api.get_proxy_metadata(id_)
            color = QColor(self.formatter.safe_format(fmt, self.mi, '', self.mi, column_name=key, template_cache=template_cache))
            color_cache[id_][key] = color
            if color.isValid():
                self.mi = None
                return color
        except Exception:
            pass


# }}}


def themed_icon_name(icon_dir, icon_name):
    root, ext = os.path.splitext(icon_name)
    # Remove any theme from the icon name
    root = root.removesuffix('-for-dark-theme').removesuffix('-for-light-theme')
    # Check if the correct themed icon exists.
    theme_suffix = '-for-dark-theme' if is_dark_theme() else '-for-light-theme'
    d = os.path.join(icon_dir, root + theme_suffix + ext)
    if os.path.exists(d):
        return d
    # No themed icon exists. Try the original name
    d = os.path.join(icon_dir, icon_name)
    return d if os.path.exists(d) else None


class ColumnIcon:  # {{{
    def __init__(self, formatter, model):
        self.mi = None
        self.formatter = formatter
        self.model = model
        self.dpr = qapplication_or_fail().devicePixelRatio()

    def __call__(self, id_, fmts, cache_index, db, icon_cache, icon_bitmap_cache, template_cache):
        if id_ in icon_cache and cache_index in icon_cache[id_]:
            self.mi = None
            return icon_cache[id_][cache_index]
        try:
            if self.mi is None:
                self.mi = db.new_api.get_proxy_metadata(id_)
            icons = []
            for dex, (kind, fmt) in enumerate(fmts):
                rule_icons = self.formatter.safe_format(fmt, self.mi, '', self.mi, column_name=cache_index + str(dex), template_cache=template_cache)
                if not rule_icons:
                    continue
                icon_list = [ic.strip() for ic in rule_icons.split(':') if ic.strip()]
                icons.extend(icon_list)
                if icon_list and not kind.endswith('_composed'):
                    break

            if icons:
                icon_string = ':'.join(icons)
                if icon_string in icon_bitmap_cache:
                    icon_bitmap = icon_bitmap_cache[icon_string]
                    icon_cache[id_][cache_index] = icon_bitmap
                    return icon_bitmap

                icon_bitmaps = []
                total_width = 0
                rh = max(2, self.model.row_height - 4)
                dim = int(self.dpr * rh)
                icon_dir = os.path.join(config_dir, 'cc_icons')
                for icon in icons:
                    d = themed_icon_name(icon_dir, icon)
                    if d is not None:
                        bm = QPixmap(d)
                        scaled, nw, nh = fit_image(bm.width(), bm.height(), bm.width(), dim)
                        bm = bm.scaled(
                            int(nw),
                            int(nh),
                            aspectRatioMode=Qt.AspectRatioMode.IgnoreAspectRatio,
                            transformMode=Qt.TransformationMode.SmoothTransformation,
                        )
                        bm.setDevicePixelRatio(self.dpr)
                        icon_bitmaps.append(bm)
                        total_width += bm.width()
                if len(icon_bitmaps) > 1:
                    i = len(icon_bitmaps)
                    result = QPixmap(total_width + ((i - 1) * 2), dim)
                    result.setDevicePixelRatio(self.dpr)
                    result.fill(Qt.GlobalColor.transparent)
                    painter = QPainter(result)
                    x = 0
                    for bm in icon_bitmaps:
                        painter.drawPixmap(x, 0, bm)
                        x += int(bm.width() / self.dpr) + 2
                    painter.end()
                else:
                    result = icon_bitmaps[0]

                icon_cache[id_][cache_index] = result
                icon_bitmap_cache[icon_string] = result
                self.mi = None
                return result
        except Exception:
            pass


# }}}


class BooksModel(QAbstractTableModel):  # {{{
    about_to_be_sorted = pyqtSignal(object, name='aboutToBeSorted')
    sorting_done = pyqtSignal(object, name='sortingDone')
    database_changed = pyqtSignal(object, name='databaseChanged')
    new_bookdisplay_data = pyqtSignal(object)
    count_changed_signal = pyqtSignal(int)
    searched = pyqtSignal(object)
    search_done = pyqtSignal()

    def __init__(self, parent=None, buffer=40):
        QAbstractTableModel.__init__(self, parent)
        base_font = parent.font() if parent else qapplication_or_fail().font()
        self.zero_page_cache = {}
        self.update_page_count_timer = t = QTimer(self)
        t.setSingleShot(True), t.setInterval(1000), t.timeout.connect(self.update_zero_page_values)
        self.bold_font = QFont(base_font)
        self.bold_font.setBold(True)
        self.italic_font = QFont(base_font)
        self.italic_font.setItalic(True)
        self.bi_font = QFont(self.bold_font)
        self.bi_font.setItalic(True)
        self.styled_columns = {}
        self.orig_headers = {
            'title': _('Title'),
            'ondevice': _('On Device'),
            'authors': _('Author(s)'),
            'size': _('Size (MB)'),
            'timestamp': _('Date'),
            'pubdate': _('Published'),
            'rating': _('Rating'),
            'publisher': _('Publisher'),
            'tags': _('Tags'),
            'series': ngettext('Series', 'Series', 1),
            'last_modified': _('Modified'),
            'languages': _('Languages'),
            'formats': _('Formats'),
            'id': _('Id'),
            'path': _('Path'),
            'pages': _('Pages'),
        }
        self.db = None

        self.formatter = SafeFormat()
        self._clear_caches()
        self.column_color = ColumnColor(self.formatter)
        self.column_icon = ColumnIcon(self.formatter, self)

        self.book_on_device = None
        self.editable_cols = [
            'title',
            'authors',
            'rating',
            'publisher',
            'tags',
            'series',
            'timestamp',
            'pubdate',
            'languages',
        ]
        self.sorted_on = DEFAULT_SORT
        self.sort_history = [self.sorted_on]
        self.last_search = ''  # The last search performed on this model
        self.column_map = []
        self.headers = {}
        self.alignment_map = {}
        self.buffer_size = buffer
        self.metadata_backup = None
        icon_height = (parent.fontMetrics() if hasattr(parent, 'fontMetrics') else QFontMetrics(QApplication.font())).lineSpacing()
        self.bool_yes_icon = QIcon.cached_icon('ok.png').pixmap(icon_height)
        self.bool_no_icon = QIcon.cached_icon('list_remove.png').pixmap(icon_height)
        self.bool_blank_icon = QIcon.cached_icon('blank.png').pixmap(icon_height)
        # Qt auto-scales marked icon correctly, so we don't need to do it (and
        # remember that the cover grid view needs a larger version of the icon,
        # anyway)
        self.marked_icon = QIcon.cached_icon('marked.png')
        self.bool_blank_icon_as_icon = QIcon.cached_icon('blank.png')
        self.row_decoration = None
        self.device_connected = False
        self.ids_to_highlight = []
        self.ids_to_highlight_set = set()
        self.current_highlighted_idx = None
        self.highlight_only = False
        self.row_height = 0
        self.marked_text_icons = {}
        self.db_prefs = {'column_color_rules': (), 'column_icon_rules': ()}
        self.read_config()

    def marked_text_icon_for(self, label):
        import random

        ans = self.marked_text_icons.get(label)
        if ans is not None:
            return ans[1]
        db = self.db
        assert db is not None
        used_labels = db.data.all_marked_labels()
        for qlabel in tuple(self.marked_text_icons):
            if qlabel not in used_labels:
                del self.marked_text_icons[qlabel]
        used_colors = {x[0] for x in self.marked_text_icons.values()}
        if qapplication_or_fail().is_dark_theme:
            all_colors = {dark_link_color, 'lightgreen', 'red', 'maroon', 'cyan', 'pink'}
        else:
            all_colors = {'blue', 'green', 'red', 'maroon', 'cyan', 'pink'}
        for c in all_colors - used_colors:
            color = c
            break
        else:
            color = random.choice(sorted(all_colors))
        pm = render_pin(color)
        ans = QIcon(pm)
        self.marked_text_icons[label] = color, ans
        return ans

    @property
    def default_image(self):
        from calibre.gui2.ui import get_gui

        gui = get_gui()
        dpr = gui.devicePixelRatio() if gui else 0
        return qapplication_or_fail().cached_qimage('default_cover.png', device_pixel_ratio=dpr)

    def _clear_caches(self):
        self.color_cache = defaultdict(dict)
        self.icon_cache = defaultdict(dict)
        self.icon_bitmap_cache = {}
        self.cover_grid_emblem_cache = defaultdict(dict)
        self.bookshelf_emblem_cache = defaultdict(dict)
        self.cover_grid_bitmap_cache = {}
        self.bookshelf_bitmap_cache = {}
        self.color_row_fmt_cache = None
        self.color_template_cache = {}
        self.icon_template_cache = {}
        self.cover_grid_template_cache = {}
        self.bookshelf_template_cache = {}

    def set_row_height(self, height):
        self.row_height = height

    def set_row_decoration(self, current_marked):
        self.row_decoration = self.bool_blank_icon_as_icon if current_marked else None

    def change_alignment(self, colname, alignment):
        if colname in self.column_map and alignment in ('left', 'right', 'center'):
            old = self.alignment_map.get(colname, 'left')
            if old == alignment:
                return
            self.alignment_map.pop(colname, None)
            if alignment != 'left':
                self.alignment_map[colname] = alignment
            col = self.column_map.index(colname)
            for row in range(self.rowCount(QModelIndex())):
                self.dataChanged.emit(self.index(row, col), self.index(row, col))

    def change_column_font(self, colname, font_type):
        if colname in self.column_map and font_type in ('normal', 'bold', 'italic', 'bi'):
            sdb = self.db
            assert sdb is not None
            db = sdb.new_api
            old = db.pref('styled_columns', {})
            old.pop(colname, None)
            self.styled_columns.pop(colname, None)
            if font_type != 'normal':
                self.styled_columns[colname] = getattr(self, f'{font_type}_font')
                old[colname] = font_type
            db.set_pref('styled_columns', old)
            col = self.column_map.index(colname)
            for row in range(self.rowCount(QModelIndex())):
                self.dataChanged.emit(self.index(row, col), self.index(row, col))

    def is_custom_column(self, cc_label):
        try:
            return cc_label in self.custom_columns
        except AttributeError:
            return False

    def read_config(self):
        pass

    def set_device_connected(self, is_connected):
        self.device_connected = is_connected

    def refresh_ondevice(self):
        db = self.db
        assert db is not None
        db.refresh_ondevice()
        self.resort()
        self.research()

    def set_book_on_device_func(self, func):
        self.book_on_device = func

    def set_database(self, db):
        self.ids_to_highlight = []
        self.zero_page_cache = {}
        if db:
            style_map = {'bold': self.bold_font, 'bi': self.bi_font, 'italic': self.italic_font}
            self.styled_columns = {k: style_map.get(v, None) for k, v in db.new_api.pref('styled_columns', {}).items()}
        self.alignment_map = {}
        self.ids_to_highlight_set = set()
        self.current_highlighted_idx = None
        self.db = db
        self.update_db_prefs_cache()
        self.custom_columns = self.db.field_metadata.custom_field_metadata()
        self.column_map = list(self.orig_headers.keys()) + list(self.custom_columns)

        assert db is not None

        def col_idx(name):
            if name == 'ondevice':
                return -1
            if name not in db.field_metadata:
                return 100000
            return db.field_metadata[name]['rec_index']

        self.column_map.sort(key=col_idx)
        for col in self.column_map:
            if col in self.orig_headers:
                self.headers[col] = self.orig_headers[col]
            elif col in self.custom_columns:
                self.headers[col] = self.custom_columns[col]['name']

        self.build_data_convertors()
        self.beginResetModel(), self.endResetModel()
        self.database_changed.emit(db)
        self.stop_metadata_backup()
        self.start_metadata_backup()

    def get_pages_description_for_zero_page(self, book_id: int) -> str:
        if (ans := self.zero_page_cache.get(book_id)) is None:
            db = self.db
            assert db is not None
            self.zero_page_cache[book_id] = ans = bool(db.new_api.pages_needs_scan((book_id,)))
        if ans:
            self.update_page_count_timer.start()
            return _('calculating')
        return '0'

    def update_zero_page_values(self) -> None:
        db = self.db
        assert db is not None
        needs_scan = tuple(bid for bid, needs_scan in self.zero_page_cache.items() if needs_scan)
        new_needs_scan = db.new_api.pages_needs_scan(needs_scan)
        changed = frozenset(needs_scan) - new_needs_scan
        if changed:
            cc = 0
            for idx, name in enumerate(self.column_map):
                if name == 'pages':
                    cc = idx
                    break
            for book_id in changed:
                del self.zero_page_cache[book_id]
                try:
                    row = db.data.id_to_index(book_id)
                except Exception:
                    pass
                else:
                    self.dataChanged.emit(self.index(row, cc), self.index(row, cc))

    def update_db_prefs_cache(self):
        db = self.db
        assert db is not None
        self.db_prefs = {
            'column_icon_rules': tuple(db.new_api.pref('column_icon_rules', ())),
            'column_color_rules': tuple(db.new_api.pref('column_color_rules', ())),
        }

    def start_metadata_backup(self):
        from calibre.db.backup import MetadataBackup

        self.metadata_backup = MetadataBackup(self.db)
        self.metadata_backup.start()

    def stop_metadata_backup(self):
        metadata_backup = getattr(self, 'metadata_backup', None)
        if metadata_backup is not None:
            metadata_backup.stop()
            # Would like to a join here, but the thread might be waiting to
            # do something on the GUI thread. Deadlock.

    def refresh_ids(self, ids, current_row=-1):
        self._clear_caches()
        db = self.db
        assert db is not None
        rows = db.refresh_ids(ids)
        if rows:
            self.refresh_rows(rows, current_row=current_row)

    def refresh_rows(self, rows, current_row=-1):
        self._clear_caches()
        cc = self.columnCount(QModelIndex()) - 1
        db = self.db
        assert db is not None
        for r in rows:
            db.new_api.clear_extra_files_cache(db.id(r))
        for first_row, last_row in group_numbers(rows):
            self.dataChanged.emit(self.index(first_row, 0), self.index(last_row, cc))
            if current_row >= 0 and first_row <= current_row <= last_row:
                self.new_bookdisplay_data.emit(self.get_book_display_info(current_row))

    def close(self):
        self.update_page_count_timer.stop()
        if self.db is not None:
            self.db.close()
            self.db = None
            self.beginResetModel(), self.endResetModel()

    def add_books(self, paths, formats, metadata, add_duplicates=False, return_ids=False):
        db = self.db
        assert db is not None
        ret = db.add_books(paths, formats, metadata, add_duplicates=add_duplicates, return_ids=return_ids)
        self.count_changed()
        return ret

    def add_news(self, path, arg):
        db = self.db
        assert db is not None
        ret = db.add_news(path, arg)
        self.count_changed()
        return ret

    def add_catalog(self, path, title):
        db = self.db
        assert db is not None
        ret = db.add_catalog(path, title)
        self.count_changed()
        return ret

    def count_changed(self, *args):
        self._clear_caches()
        db = self.db
        assert db is not None
        self.count_changed_signal.emit(db.count())

    def counts(self):
        db = self.db
        assert db is not None
        library_total = total = db.count()
        if db.data.search_restriction_applied():
            total = db.data.get_search_restriction_book_count()
        return Counts(library_total, total, self.count())

    def row_indices(self, index):
        """Return list indices of all cells in index.row()"""
        return [self.index(index.row(), c) for c in range(self.columnCount())]

    @property
    def by_author(self):
        return self.sorted_on[0] == 'authors'

    def books_deleted(self):
        self.count_changed()
        self.beginResetModel(), self.endResetModel()

    def delete_books(self, indices, permanent=False):
        ids = list(map(self.id, indices))
        self.delete_books_by_id(ids, permanent=permanent)
        return ids

    def delete_books_by_id(self, ids, permanent=False):
        db = self.db
        assert db is not None
        db.new_api.remove_books(ids, permanent=permanent)
        self.ids_deleted(ids)

    def ids_deleted(self, ids):
        db = self.db
        assert db is not None
        db.data.books_deleted(tuple(ids))
        db.notify('delete', list(ids))
        self.books_deleted()

    def books_added(self, num):
        if num > 0:
            self.beginInsertRows(QModelIndex(), 0, num - 1)
            self.endInsertRows()
            self.count_changed()

    def set_highlight_only(self, toWhat):
        self.highlight_only = toWhat

    def get_current_highlighted_id(self):
        if len(self.ids_to_highlight) == 0 or self.current_highlighted_idx is None:
            return None
        try:
            return self.ids_to_highlight[self.current_highlighted_idx]
        except Exception:
            return None

    def get_next_highlighted_id(self, current_row, forward):
        if len(self.ids_to_highlight) == 0 or self.current_highlighted_idx is None:
            return None
        if current_row is None:
            row_ = self.current_highlighted_idx
        else:
            row_ = current_row
        while True:
            row_ += 1 if forward else -1
            if row_ < 0:
                row_ = self.count() - 1
            elif row_ >= self.count():
                row_ = 0
            if self.id(row_) in self.ids_to_highlight_set:
                break
        try:
            self.current_highlighted_idx = self.ids_to_highlight.index(self.id(row_))
        except Exception:
            # This shouldn't happen ...
            return None
        return self.get_current_highlighted_id()

    def highlight_ids(self, ids_to_highlight):
        self.ids_to_highlight = ids_to_highlight
        self.ids_to_highlight_set = set(self.ids_to_highlight)
        if self.ids_to_highlight:
            self.current_highlighted_idx = 0
        else:
            self.current_highlighted_idx = None
        self.beginResetModel(), self.endResetModel()

    def search(self, text, reset=True):
        db = self.db
        assert db is not None
        try:
            if self.highlight_only:
                db.search('')
                if not text:
                    self.ids_to_highlight = []
                    self.ids_to_highlight_set = set()
                    self.current_highlighted_idx = None
                else:
                    self.ids_to_highlight = db.search(text, return_matches=True)
                    self.ids_to_highlight_set = set(self.ids_to_highlight)
                    if self.ids_to_highlight:
                        self.current_highlighted_idx = 0
                    else:
                        self.current_highlighted_idx = None
            else:
                self.ids_to_highlight = []
                self.ids_to_highlight_set = set()
                self.current_highlighted_idx = None
                db.search(text)
        except ParseException as e:
            self.searched.emit(e.msg)
            return
        self.last_search = text
        if reset:
            self.beginResetModel(), self.endResetModel()
        if self.last_search:
            # Do not issue search done for the null search. It is used to clear
            # the search and count records for restrictions
            self.searched.emit(True)
        self.search_done.emit()

    def sort(self, column, order=Qt.SortOrder.AscendingOrder, reset=True):
        if not self.db:
            return
        if not isinstance(order, bool):
            order = order == Qt.SortOrder.AscendingOrder
        label = self.column_map[column]
        self._sort(label, order, reset)

    def sort_by_named_field(self, field, order, reset=True):
        db = self.db
        assert db is not None
        if field in list(db.field_metadata.keys()):
            self._sort(field, order, reset)

    def _sort(self, label, order, reset):
        db = self.db
        assert db is not None
        self.about_to_be_sorted.emit(db.id)
        db.data.incremental_sort([(label, order)])
        if reset:
            self.beginResetModel(), self.endResetModel()
        self.sorted_on = (label, order)
        self.sort_history.insert(0, self.sorted_on)
        self.sorting_done.emit(db.index)

    def refresh(self, reset=True):
        db = self.db
        assert db is not None
        db.refresh(field=None)
        self.resort(reset=reset)

    def beginResetModel(self):
        self._clear_caches()
        QAbstractTableModel.beginResetModel(self)

    def reset(self):
        self.beginResetModel(), self.endResetModel()

    def resort(self, reset=True):
        if not self.db:
            return
        self.db.multisort(self.sort_history[: tweaks['maximum_resort_levels']])
        if self.sort_history:
            # There are ways to get here that don't set sorted_on, e.g., the
            # sort_by action. I don't think sort_history can be empty, but it
            # doesn't hurt to check
            self.sorted_on = self.sort_history[0]
        if reset:
            self.beginResetModel(), self.endResetModel()

    def research(self, reset=True):
        self.search(self.last_search, reset=reset)

    def columnCount(self, parent: QModelIndex = QModelIndex()):
        if parent is not None and parent.isValid():
            return 0
        return len(self.column_map)

    def rowCount(self, parent: QModelIndex = QModelIndex()):
        if parent is not None and parent.isValid():
            return 0
        return len(self.db.data) if self.db else 0

    def all_current_book_ids(self):
        db = self.db
        assert db is not None
        return db.data._map_filtered

    def count(self):
        return self.rowCount()

    def get_book_display_info(self, idx):
        db = self.db
        assert db is not None
        mi = db.get_metadata(idx)
        mi.size = mi._proxy_metadata.book_size
        mi.book_display_info_is_from_device = False
        mi.cover_data = ('jpg', self.cover(idx))
        mi.id = db.id(idx)
        mi.field_metadata = db.field_metadata
        mi.path = db.abspath(idx, create_dirs=False)
        mi.format_files = db.new_api.format_files(db.data.index_to_id(idx))
        mi.row_number = idx
        try:
            mi.marked = db.data.get_marked(idx, index_is_id=False)
        except Exception:
            mi.marked = None
        return mi

    def current_changed(self, current, previous, emit_signal=True):
        if current.isValid():
            idx = current.row()
            db = self.db
            assert db is not None
            try:
                db.id(idx)
            except Exception:
                # can happen if an out of band search is done causing the index
                # to no longer be valid since this function is now called after
                # an event loop tick.
                return
            try:
                data = self.get_book_display_info(idx)
            except Exception as e:
                if sys.excepthook is simple_excepthook or sys.excepthook is sys.__excepthook__:
                    return  # ignore failures during startup/shutdown
                setattr(e, 'locking_violation_msg', _('Failed to read cover file for this book from the calibre library.'))
                raise
            else:
                if emit_signal:
                    self.new_bookdisplay_data.emit(data)
                else:
                    return data

    def get_book_info(self, index):
        if isinstance(index, numbers.Integral):
            index = self.index(index, 0)
        # If index is not valid returns None
        data = self.current_changed(index, None, False)
        return data

    def metadata_for(self, ids, get_cover=True):
        """
        WARNING: if get_cover=True temp files are created for mi.cover.
        Remember to delete them once you are done with them.
        """
        db = self.db
        assert db is not None
        ans = []
        for id in ids:
            mi = db.get_metadata(id, index_is_id=True, get_cover=get_cover)
            ans.append(mi)
        return ans

    def get_metadata(self, rows, rows_are_ids=False, full_metadata=False):
        db = self.db
        assert db is not None
        metadata, _full_metadata = [], []
        if not rows_are_ids:
            rows = [db.id(row.row()) for row in rows]
        for id in rows:
            mi = db.get_metadata(id, index_is_id=True)
            _full_metadata.append(mi)
            au = authors_to_string(mi.authors or [_('Unknown')])
            tags = mi.tags or []
            if mi.series is not None:
                tags.append(mi.series)
            info = {
                'title': mi.title,
                'authors': au,
                'author_sort': mi.author_sort,
                'cover': db.cover(id, index_is_id=True),
                'tags': tags,
                'comments': mi.comments,
            }
            if mi.series is not None:
                info['tag order'] = {mi.series: db.books_in_series_of(id, index_is_id=True)}

            metadata.append(info)
        if full_metadata:
            return metadata, _full_metadata
        else:
            return metadata

    def get_preferred_formats_from_ids(
        self,
        ids,
        formats,
        set_metadata=False,
        specific_format=None,
        exclude_auto=False,
        mode='r+b',
        use_plugboard=None,
        plugboard_formats=None,
        modified_metadata=None,
    ):
        from calibre.ebooks.metadata.meta import set_metadata as _set_metadata

        db = self.db
        assert db is not None
        ans = []
        need_auto = []
        modified_metadata = [] if modified_metadata is None else modified_metadata
        if specific_format is not None:
            formats = [specific_format.lower()]
        for id in ids:
            format = None
            fmts = db.formats(id, index_is_id=True)
            if not fmts:
                fmts = ''
            db_formats = set(fmts.lower().split(','))
            available_formats = {f.lower() for f in formats}
            u = available_formats.intersection(db_formats)
            for f in formats:
                if f.lower() in u:
                    format = f
                    break
            if format is not None:
                pt = PersistentTemporaryFile(suffix='caltmpfmt.' + format)
                db.copy_format_to(id, format, pt, index_is_id=True)
                pt.seek(0)
                newmi = None
                if set_metadata:
                    try:
                        mi = db.get_metadata(id, get_cover=True, index_is_id=True, cover_as_data=True)
                        if use_plugboard and plugboard_formats is not None and format.lower() in plugboard_formats:
                            plugboards = db.new_api.pref('plugboards', {})
                            cpb = find_plugboard(use_plugboard, format.lower(), plugboards)
                            if cpb:
                                newmi = mi.deepcopy_metadata()
                                newmi.template_to_attribute(mi, cpb)
                        if newmi is not None:
                            _set_metadata(pt, newmi, format)
                        else:
                            _set_metadata(pt, mi, format)
                    except Exception:
                        traceback.print_exc()
                pt.close()

                def to_uni(x):
                    if isinstance(x, bytes):
                        x = x.decode(filesystem_encoding)
                    return x

                ans.append(to_uni(os.path.abspath(pt.name)))
                modified_metadata.append(newmi)
            else:
                need_auto.append(id)
                if not exclude_auto:
                    ans.append(None)
                    modified_metadata.append(None)
        return ans, need_auto

    def get_preferred_formats(self, rows, formats, paths=False, set_metadata=False, specific_format=None, exclude_auto=False):
        from calibre.ebooks.metadata.meta import set_metadata as _set_metadata

        db = self.db
        assert db is not None
        ans = []
        need_auto = []
        if specific_format is not None:
            formats = [specific_format.lower()]
        for row in (row.row() for row in rows):
            format = None
            fmts = db.formats(row)
            if not fmts:
                fmts = ''
            db_formats = set(fmts.lower().split(','))
            available_formats = {f.lower() for f in formats}
            u = available_formats.intersection(db_formats)
            for f in formats:
                if f.lower() in u:
                    format = f
                    break
            if format is not None:
                pt = PersistentTemporaryFile(suffix='.' + format)
                db.copy_format_to(id, format, pt, index_is_id=True)
                pt.seek(0)
                if set_metadata:
                    _set_metadata(pt, db.get_metadata(row, get_cover=True, cover_as_data=True), format)
                pt.close() if paths else pt.seek(0)
                ans.append(pt)
            else:
                need_auto.append(row)
                if not exclude_auto:
                    ans.append(None)
        return ans, need_auto

    def id(self, row):
        db = self.db
        assert db is not None
        return db.id(getattr(row, 'row', lambda: row)())

    def authors(self, row_number):
        db = self.db
        assert db is not None
        return db.authors(row_number)

    def title(self, row_number):
        db = self.db
        assert db is not None
        return db.title(row_number)

    def rating(self, row_number):
        db = self.db
        assert db is not None
        ans = db.rating(row_number)
        ans = ans / 2 if ans else 0
        return int(ans)

    def cover(self, row_number):
        db = self.db
        assert db is not None
        data = None
        try:
            data = db.cover(row_number)
        except IndexError:  # Happens if database has not yet been refreshed
            pass
        except MemoryError:
            raise ValueError(_('The cover for the book %s is too large, cannot load it. Resize or delete it.') % db.title(row_number))

        if not data:
            return self.default_image
        img = QImage()
        load_from_data_without_gil(img, data)
        return self.default_image if img.isNull() else img

    def build_data_convertors(self):
        rating_fields = {}
        bool_fields = set()

        def renderer(field, decorator=False):
            db = self.db
            assert db is not None
            idfunc = db.id
            if field == 'id':

                def func(idx):
                    return idfunc(idx)

                return func
            fffunc = db.new_api.fast_field_for
            field_obj = db.new_api.fields[field]
            m = field_obj.metadata.copy()
            if 'display' not in m:
                m['display'] = {}
            dt = m['datatype']

            if decorator == 'bool':
                bool_fields.add(field)
                bt = db.new_api.pref('bools_are_tristate')
                bn = self.bool_no_icon
                by = self.bool_yes_icon

                if dt != 'bool':

                    def func(idx):
                        val = fffunc(field_obj, idfunc(idx))
                        if val is None:
                            return None
                        val = force_to_bool(val)
                        if val is None:
                            return None
                        return by if val else bn

                elif m['display'].get('bools_show_icons', True):

                    def func(idx):
                        val = force_to_bool(fffunc(field_obj, idfunc(idx)))
                        if val is None:
                            return None if bt else bn
                        return by if val else bn

                else:

                    def func(idx):
                        return None

            elif field == 'size':
                sz_mult = 1 / (1024**2)

                def func(idx):
                    val = fffunc(field_obj, idfunc(idx), default_value=0) or 0
                    if val == 0:
                        return None
                    ans = f'{val * sz_mult:.1f}'
                    return '<0.1' if ans == '0.0' else ans

            elif field == 'languages':

                def func(idx):
                    return ', '.join(calibre_langcode_to_name(x) for x in fffunc(field_obj, idfunc(idx)))

            elif field == 'pages':

                def func(idx):
                    book_id = idfunc(idx)
                    ans = fffunc(field_obj, book_id, 0)
                    match ans:
                        case 0:
                            return self.get_pages_description_for_zero_page(book_id)
                        case -1:
                            return _('None')
                        case -2:
                            return _('Error')
                        case -3:
                            return _('DRM')
                        case _:
                            return str(ans)

            elif field == 'ondevice' and decorator:
                by = self.bool_yes_icon
                bb = self.bool_blank_icon

                def func(idx):
                    return by if fffunc(field_obj, idfunc(idx)) else bb

            elif dt in {'text', 'comments', 'composite', 'enumeration'}:
                if m['is_multiple']:
                    jv = m['is_multiple']['list_to_ui']
                    do_sort = '&' not in jv
                    if field_obj.is_composite:
                        if do_sort:
                            sv = m['is_multiple']['cache_to_list']

                            def func(idx):
                                val = fffunc(field_obj, idfunc(idx), default_value='') or ''
                                return jv.join(sorted((x.strip() for x in val.split(sv)), key=sort_key))

                        else:

                            def func(idx):
                                return fffunc(field_obj, idfunc(idx), default_value='')

                    elif do_sort:

                        def func(idx):
                            return jv.join(sorted(fffunc(field_obj, idfunc(idx), default_value=()), key=sort_key))

                    else:

                        def func(idx):
                            return jv.join(fffunc(field_obj, idfunc(idx), default_value=()))

                elif dt in {'text', 'composite', 'enumeration'} and m['display'].get('use_decorations', False):

                    def func(idx):
                        text = fffunc(field_obj, idfunc(idx))
                        return (text) if force_to_bool(text) is None else None

                else:

                    def func(idx):
                        return fffunc(field_obj, idfunc(idx), default_value='')

            elif dt == 'datetime':

                def func(idx):
                    val = fffunc(field_obj, idfunc(idx), default_value=UNDEFINED_DATE)
                    return None if is_date_undefined(val) else qt_from_dt(val)

            elif dt == 'rating':
                rating_fields[field] = m['display'].get('allow_half_stars', False)

                def func(idx):
                    return int(fffunc(field_obj, idfunc(idx), default_value=0))

            elif dt == 'series':
                sidx_field = db.new_api.fields[field + '_index']

                def func(idx):
                    book_id = idfunc(idx)
                    series = fffunc(field_obj, book_id, default_value=False)
                    if series:
                        return f'{series} [{fmt_sidx(fffunc(sidx_field, book_id, default_value=1.0))}]'
                    return None

            elif dt in {'int', 'float'}:
                fmt = m['display'].get('number_format', None)

                def func(idx):
                    val = fffunc(field_obj, idfunc(idx))
                    if val is None:
                        return None
                    if fmt:
                        try:
                            return fmt.format(val)
                        except TypeError, ValueError, AttributeError, IndexError, KeyError:
                            pass
                    return val

            elif dt == 'bool':
                if m['display'].get('bools_show_text', False):

                    def func(idx):
                        v = fffunc(field_obj, idfunc(idx))
                        return None if v is None else (_('Yes') if v else _('No'))

                else:

                    def func(idx):
                        return None

            else:

                def func(idx):
                    return None

            return func

        self.dc = {f: renderer(f) for f in self.orig_headers.keys()}
        self.dc_decorator = {f: renderer(f, True) for f in ('ondevice',)}

        for col in self.custom_columns:
            self.dc[col] = renderer(col)
            m = self.custom_columns[col]
            dt = m['datatype']
            mult = m['is_multiple']
            if dt in {'text', 'composite', 'enumeration'} and not mult and m['display'].get('use_decorations', False):
                self.dc_decorator[col] = renderer(col, 'bool')
            elif dt == 'bool':
                self.dc_decorator[col] = renderer(col, 'bool')

        tc = self.dc.copy()

        def stars_tooltip(func, allow_half=True):
            def f(idx):
                val = int(func(idx))
                ans = str(val // 2)
                if allow_half and val % 2:
                    ans += '.5'
                return _('%s stars') % ans

            return f

        def bool_tooltip(key):
            def f(idx):
                db = self.db
                assert db is not None
                return db.new_api.fast_field_for(db.new_api.fields[key], db.id(idx))

            return f

        def template_tooltip(key, orig_tt_func):
            def f(idx):
                try:
                    db = self.db
                    assert db is not None
                    template = db.new_api.pref('column_tooltip_templates', {}).get(key, '')
                    # Simulate what Qt does to get the text for non-string data
                    # types so that idx.data(Qt.ItemDataRole.ToolTipRole)
                    # always returns a string.
                    orig_obj = orig_tt_func(idx)
                    match orig_obj:
                        case None:
                            v = ''
                        case str():
                            v = orig_obj
                        case bool():
                            v = str(orig_obj)
                        case int() | float() | QDateTime():
                            v = QLocale().toString(orig_obj)
                        case _:
                            v = f'unsupported type {type(orig_obj)}'
                    if template:
                        global_vars = {'column_lookup_name': key, 'original_text': v}
                        mi = db.new_api.get_proxy_metadata(db.data.index_to_id(idx))
                        return self.formatter.safe_format(template, {}, _('tooltip template error'), mi, global_vars=global_vars)
                    return v
                except Exception as e:
                    return str(e)

            return f

        for f, allow_half in rating_fields.items():
            tc[f] = stars_tooltip(self.dc[f], allow_half)
        for f in bool_fields:
            tc[f] = bool_tooltip(f)
        # Make all columns use the template tooltip renderer. This avoids the
        # "first time" problem where a template tooltip is added but not
        # used until calibre is restarted
        for f in self.column_map:
            tc[f] = template_tooltip(f, tc[f])
        # build a index column to data converter map, to remove the string lookup in the data loop
        self.column_to_dc_map = [self.dc[col] for col in self.column_map]
        self.column_to_tc_map = [tc[col] for col in self.column_map]
        self.column_to_dc_decorator_map = [self.dc_decorator.get(col, None) for col in self.column_map]

    def data(self, index, role=...):
        db = self.db
        assert db is not None
        if db.new_api.is_doing_rebuild_or_vacuum:
            return None
        col = index.column()
        # in obscure cases where custom columns are both edited and added, for a time
        # the column map does not accurately represent the screen. In these cases,
        # we will get asked to display columns we don't know about. Must test for this.
        if col >= len(self.column_to_dc_map) or col < 0:
            return None
        try:
            return self.actual_data(index, role, col)
        except Exception:
            import traceback

            traceback.print_exc()

    def actual_data(self, index, role, col):
        if role == Qt.ItemDataRole.DisplayRole:
            rules = self.db_prefs['column_icon_rules']
            if rules:
                key = self.column_map[col]
                id_ = None
                fmts = []
                for kind, k, fmt in rules:
                    if k == key and kind in {'icon_only', 'icon_only_composed'}:
                        if id_ is None:
                            id_ = self.id(index)
                            self.column_icon.mi = None
                        fmts.append((kind, fmt))

                if fmts:
                    cache_index = key + ':DisplayRole'
                    ccicon = self.column_icon(
                        id_,
                        fmts,
                        cache_index,
                        self.db,
                        self.icon_cache,
                        self.icon_bitmap_cache,
                        self.icon_template_cache,
                    )
                    if ccicon is not None:
                        return None
                    self.icon_cache[id_][cache_index] = None
            return self.column_to_dc_map[col](index.row())
        elif role == Qt.ItemDataRole.ToolTipRole:
            return self.column_to_tc_map[col](index.row())
        elif role == Qt.ItemDataRole.EditRole:
            return self.column_to_dc_map[col](index.row())
        elif role == Qt.ItemDataRole.BackgroundRole:
            if self.id(index) in self.ids_to_highlight_set:
                return QColor('#027524') if qapplication_or_fail().is_dark_theme else QColor('#b4ecb4')
        elif role == Qt.ItemDataRole.ForegroundRole:
            key = self.column_map[col]
            id_ = self.id(index)
            self.column_color.mi = None

            for k, fmt in self.db_prefs['column_color_rules']:
                if k == key:
                    ccol = self.column_color(id_, key, fmt, self.db, self.color_cache, self.color_template_cache)
                    if ccol is not None:
                        return ccol

            if self.is_custom_column(key) and self.custom_columns[key]['datatype'] == 'enumeration':
                cc = self.custom_columns[self.column_map[col]]['display']
                colors = cc.get('enum_colors', [])
                values = cc.get('enum_values', [])
                txt = str(index.data(Qt.ItemDataRole.DisplayRole) or '')
                if len(colors) > 0 and txt in values:
                    try:
                        color = QColor(colors[values.index(txt)])
                        if color.isValid():
                            self.column_color.mi = None
                            return color
                    except Exception:
                        pass

            if self.color_row_fmt_cache is None:
                self.color_row_fmt_cache = tuple(fmt for key, fmt in self.db_prefs['column_color_rules'] if key == color_row_key)
            for fmt in self.color_row_fmt_cache:
                ccol = self.column_color(id_, color_row_key, fmt, self.db, self.color_cache, self.color_template_cache)
                if ccol is not None:
                    return ccol

            self.column_color.mi = None
            return None
        elif role == Qt.ItemDataRole.DecorationRole:
            default_icon = None
            if self.column_to_dc_decorator_map[col] is not None:
                default_icon = self.column_to_dc_decorator_map[index.column()](index.row())
            rules = self.db_prefs['column_icon_rules']
            if rules:
                key = self.column_map[col]
                id_ = None
                need_icon_with_text = False
                fmts = []
                for kind, k, fmt in rules:
                    if k == key and kind.startswith('icon'):
                        if id_ is None:
                            id_ = self.id(index)
                            self.column_icon.mi = None
                        fmts.append((kind, fmt))
                        if kind in ('icon', 'icon_composed'):
                            need_icon_with_text = True
                if fmts:
                    cache_index = key + ':DecorationRole'
                    ccicon = self.column_icon(
                        id_,
                        fmts,
                        cache_index,
                        self.db,
                        self.icon_cache,
                        self.icon_bitmap_cache,
                        self.icon_template_cache,
                    )
                    if ccicon is not None:
                        return ccicon
                    if need_icon_with_text and default_icon is None:
                        self.icon_cache[id_][cache_index] = self.bool_blank_icon
                        return self.bool_blank_icon
                    self.icon_cache[id_][cache_index] = None
            return default_icon
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            cname = self.column_map[index.column()]
            ans = Qt.AlignmentFlag.AlignVCenter | ALIGNMENT_MAP[self.alignment_map.get(cname, 'left')]
            return int(ans)  # https://bugreports.qt.io/browse/PYSIDE-1974
        elif role == Qt.ItemDataRole.FontRole and self.styled_columns:
            cname = self.column_map[index.column()]
            return self.styled_columns.get(cname)
        elif role == Qt.ItemDataRole.UserRole:
            return self.id(index)
        # elif role == Qt.ItemDataRole.ToolTipRole and index.isValid():
        #     if self.column_map[index.column()] in self.editable_cols:
        #         return (_("Double click to <b>edit</b> me<br><br>"))
        return None

    def headerData(self, section, orientation, role=...):
        db = self.db
        if db is None:
            return None
        if orientation == Qt.Orientation.Horizontal:
            if section >= len(self.column_map):  # same problem as in data, the column_map can be wrong
                return None
            if role == Qt.ItemDataRole.ToolTipRole:
                ht = self.column_map[section]
                title = self.headers[ht]
                fm = db.field_metadata[self.column_map[section]]
                if ht == 'timestamp':  # change help text because users know this field as 'date'
                    ht = 'date'
                if fm['is_category']:
                    is_cat = '<br><br>' + prepare_string_for_xml(_('Click in this column and press Q to Quickview books with the same "%s"') % ht)
                else:
                    is_cat = ''
                cust_desc = ''
                if fm['is_custom']:
                    cust_desc = fm['display'].get('description', '')
                    if cust_desc:
                        cust_desc = (
                            '<br><b>{}</b>'.format(_('Description:')) + '<span style="white-space:pre-wrap"> ' + prepare_string_for_xml(cust_desc) + '</span>'
                        )
                return '<b>{}</b>: {}'.format(
                    prepare_string_for_xml(title),
                    _('The lookup/search name is <i>{0}</i>').format(ht) + cust_desc + is_cat,
                )
            if role == Qt.ItemDataRole.DisplayRole:
                return self.headers[self.column_map[section]]
            return None
        if role == Qt.ItemDataRole.ToolTipRole and orientation == Qt.Orientation.Vertical:
            col = db.field_metadata['marked']['rec_index']
            marked = db.data[section][col]
            if marked is not None:
                s = _('Marked with text "{0}"').format(marked) if marked != 'true' else _('Marked without text')
            else:
                s = ''
            if DEBUG:
                col = db.field_metadata['uuid']['rec_index']
                s += ('\n' if s else '') + _('This book\'s UUID is "{0}"').format(db.data[section][col])
            return s

        if role == Qt.ItemDataRole.DisplayRole:  # orientation is vertical
            return section + 1
        if role == Qt.ItemDataRole.DecorationRole:
            try:
                m = db.data.get_marked(db.data.index_to_id(section))
                if m:
                    i = self.marked_icon if m == 'true' else self.marked_text_icon_for(m)
                else:
                    i = self.row_decoration
                return i
            except ValueError, IndexError:
                pass
        return None

    def flags(self, index):
        flags = QAbstractTableModel.flags(self, index)
        if index.isValid():
            colhead = self.column_map[index.column()]
            if colhead in self.editable_cols:
                flags |= Qt.ItemFlag.ItemIsEditable
            elif self.is_custom_column(colhead):
                if self.custom_columns[colhead]['is_editable']:
                    flags |= Qt.ItemFlag.ItemIsEditable
        return flags

    def set_custom_column_data(self, row, colhead, value):
        db = self.db
        assert db is not None
        cc = self.custom_columns[colhead]
        typ = cc['datatype']
        label = db.field_metadata.key_to_label(colhead)
        s_index = None
        if typ in ('text', 'comments'):
            val = str(value or '').strip()
            val = val or None
        elif typ == 'enumeration':
            val = str(value or '').strip()
            if not val:
                val = None
        elif typ == 'bool':
            val = value if value is None else bool(value)
        elif typ == 'rating':
            val = max(0, min(int(value or 0), 10))
        elif typ in ('int', 'float'):
            if value == 0:
                val = '0'
            else:
                val = str(value or '').strip()
            if not val:
                val = None
        elif typ == 'datetime':
            val = value
            if val is None:
                val = None
            else:
                if not val.isValid():
                    return False
                val = qt_to_dt(val, as_utc=False)
        elif typ == 'series':
            val = str(value or '').strip()
            if val:
                pat = re.compile(r'\[([.0-9]+)\]')
                match = pat.search(val)
                if match is not None:
                    s_index = float(match.group(1))
                    val = pat.sub('', val).strip()
                elif val:
                    # it is OK to leave s_index == None when using 'no_change'
                    if tweaks['series_index_auto_increment'] != 'const' and tweaks['series_index_auto_increment'] != 'no_change':
                        s_index = db.get_next_cc_series_num_for(val, label=label, num=None)
        elif typ == 'composite':
            tmpl = str(value or '').strip()
            disp = cc['display']
            disp['composite_template'] = tmpl
            db.set_custom_column_metadata(cc['colnum'], display=disp, update_last_modified=True)
            self.refresh(reset=False)
            self.research(reset=True)
            return True

        id = db.id(row)
        books_to_refresh = {id}
        books_to_refresh |= db.set_custom(id, val, extra=s_index, label=label, num=None, append=False, notify=True, allow_case_change=True)
        self.refresh_ids(list(books_to_refresh), current_row=row)
        return True

    def setData(self, index, value, role=...):
        from calibre.gui2.ui import get_gui

        gui = get_gui()
        if gui is not None and gui.shutting_down:
            return False
        if role == Qt.ItemDataRole.EditRole:
            from calibre.gui2.ui import get_gui

            try:
                return self._set_data(index, value)
            except OSError as err:
                import traceback

                traceback.print_exc()
                det_msg = traceback.format_exc()
                gui = get_gui()
                assert gui is not None
                if gui.show_possible_sharing_violation(err, det_msg):
                    return False
                error_dialog(
                    get_gui(),
                    _('Failed to set data'),
                    _('Could not set data, click "Show details" to see why.'),
                    det_msg=traceback.format_exc(),
                    show=True,
                )
            except Exception:
                import traceback

                traceback.print_exc()
                error_dialog(
                    get_gui(),
                    _('Failed to set data'),
                    _('Could not set data, click "Show details" to see why.'),
                    det_msg=traceback.format_exc(),
                    show=True,
                )
        return False

    def _set_data(self, index, value):
        db = self.db
        assert db is not None
        row, col = index.row(), index.column()
        column = self.column_map[col]
        if self.is_custom_column(column):
            if not self.set_custom_column_data(row, column, value):
                return False
        else:
            if column not in self.editable_cols:
                return False
            val = int(value) if column == 'rating' else value if column in ('timestamp', 'pubdate') else re.sub(r'\s', ' ', str(value or '').strip())
            id = db.id(row)
            books_to_refresh = {id}
            if column == 'rating':
                val = max(0, min(int(val or 0), 10))
                db.set_rating(id, val)
            elif column == 'series':
                val = str(val).strip()
                if not val:
                    books_to_refresh |= db.set_series(id, val, allow_case_change=True)
                    db.set_series_index(id, 1.0)
                else:
                    pat = re.compile(r'\[([.0-9]+)\]')
                    match = pat.search(val)
                    if match is not None:
                        db.set_series_index(id, float(match.group(1)))
                        val = pat.sub('', val).strip()
                    elif val:
                        if tweaks['series_index_auto_increment'] != 'const' and tweaks['series_index_auto_increment'] != 'no_change':
                            ni = db.get_next_series_num_for(val)
                            if ni != 1:
                                db.set_series_index(id, ni)
                    if val:
                        books_to_refresh |= db.set_series(id, val, allow_case_change=True)
            elif column == 'timestamp':
                if not isinstance(val, QDateTime) or not val.isValid():
                    return False
                db.set_timestamp(id, qt_to_dt(val, as_utc=False))
            elif column == 'pubdate':
                if not isinstance(val, QDateTime) or not val.isValid():
                    return False
                db.set_pubdate(id, qt_to_dt(val, as_utc=False))
            elif column == 'languages':
                val = str(val).split(',')
                db.set_languages(id, val)
            else:
                if column == 'authors' and val:
                    val = authors_to_string(string_to_authors(val))
                books_to_refresh |= db.set(row, column, val, allow_case_change=True)
                if column == 'title':
                    if (lang := db.languages(row)) is not None:
                        lang = lang.partition(',')[0]
                    db.set(row, 'sort', title_sort(val, lang=lang), allow_case_change=True)
            self.refresh_ids(list(books_to_refresh), row)
        self.dataChanged.emit(index, index)
        return True


# }}}


class OnDeviceSearch(SearchQueryParser):  # {{{
    USABLE_LOCATIONS = [
        'all',
        'author',
        'authors',
        'collections',
        'format',
        'formats',
        'title',
        'inlibrary',
        'tags',
        'search',
    ]

    def __init__(self, model):
        SearchQueryParser.__init__(self, locations=self.USABLE_LOCATIONS)
        self.model = model

    def universal_set(self):
        return set(range(len(self.model.db)))

    def get_matches(self, location, query, candidates=None):
        location = location.lower().strip()
        if location == 'authors':
            location = 'author'

        matchkind = CONTAINS_MATCH
        if len(query) > 1:
            if query.startswith('\\'):
                query = query[1:]
            elif query.startswith('='):
                matchkind = EQUALS_MATCH
                query = query[1:]
            elif query.startswith('~'):
                matchkind = REGEXP_MATCH
                query = query[1:]
        if matchkind != REGEXP_MATCH:  # leave case in regexps because it can be significant e.g. \S \W \D
            query = query.lower()

        if location not in self.USABLE_LOCATIONS:
            return set()
        matches = set()
        all_locs = set(self.USABLE_LOCATIONS) - {'all', 'tags', 'search'}
        locations = all_locs if location == 'all' else [location]
        q = {
            'title': lambda x: getattr(x, 'title').lower(),
            'author': lambda x: ' & '.join(getattr(x, 'authors')).lower(),
            'collections': lambda x: ','.join(getattr(x, 'device_collections')).lower(),
            'format': lambda x: os.path.splitext(x.path)[1].lower(),
            'inlibrary': lambda x: getattr(x, 'in_library'),
            'tags': lambda x: getattr(x, 'tags', []),
        }
        for x in ('author', 'format'):
            q[x + 's'] = q[x]
        upf = prefs['use_primary_find_in_search']
        for index, row in enumerate(self.model.db):
            for locvalue in locations:
                accessor = q[locvalue]
                if query == 'true':
                    if accessor(row):
                        matches.add(index)
                    continue
                if query == 'false':
                    if not accessor(row):
                        matches.add(index)
                    continue
                if locvalue == 'inlibrary':
                    continue  # this is bool, so can't match below
                try:
                    # Can't separate authors because comma is used for name sep and author sep
                    # Exact match might not get what you want. For that reason, turn author
                    # exactmatch searches into contains searches.
                    if locvalue == 'author' and matchkind == EQUALS_MATCH:
                        m = CONTAINS_MATCH
                    else:
                        m = matchkind

                    vals = accessor(row)
                    if vals is None:
                        vals = ''
                    if isinstance(vals, (str, bytes)):
                        vals = vals.split(',') if locvalue == 'collections' else [vals]
                    if _match(query, vals, m, use_primary_find_in_search=upf):
                        matches.add(index)
                        break
                except ValueError:  # Unicode errors
                    traceback.print_exc()
        return matches


# }}}


class DeviceDBSortKeyGen:  # {{{
    def __init__(self, attr, keyfunc, db):
        self.attr = attr
        self.db = db
        self.keyfunc = keyfunc

    def __call__(self, x) -> bytes | str:
        try:
            ans = self.keyfunc(getattr(self.db[x], self.attr))
        except Exception:
            ans = ''
        return ans


# }}}


class DeviceBooksModel(BooksModel):  # {{{
    booklist_dirtied = pyqtSignal()
    upload_collections = pyqtSignal(object)
    resize_rows = pyqtSignal()
    db: BookList
    map: list[int]
    sorted_map: list[int]

    def __init__(self, parent):
        BooksModel.__init__(self, parent)
        self.db = BookList()
        self.map = []
        self.sorted_map = []
        self.sorted_on = DEFAULT_SORT
        self.sort_history = [self.sorted_on]
        self.unknown = _('Unknown')
        self.column_map = ['inlibrary', 'title', 'authors', 'timestamp', 'size', 'collections']
        self.headers = {
            'inlibrary': _('In Library'),
            'title': _('Title'),
            'authors': _('Author(s)'),
            'timestamp': _('Date'),
            'size': _('Size'),
            'collections': _('Collections'),
        }
        self.marked_for_deletion = {}
        self.search_engine = OnDeviceSearch(self)
        self.editable = ['title', 'authors', 'collections']
        self.book_in_library = None
        self.sync_icon = QIcon.cached_icon('sync.png')

    def counts(self):
        return Counts(len(self.db), len(self.db), len(self.map))

    def count_changed(self, *args):
        self.count_changed_signal.emit(len(self.db))

    def mark_for_deletion(self, job, rows, rows_are_ids=False):
        db_indices = rows if rows_are_ids else self.indices(rows)
        db_items = [self.db[i] for i in db_indices if -1 < i < len(self.db)]
        self.marked_for_deletion[job] = db_items
        if rows_are_ids:
            self.beginResetModel(), self.endResetModel()
        else:
            for row in rows:
                indices = self.row_indices(row)
                self.dataChanged.emit(indices[0], indices[-1])

    def find_item_in_db(self, item):
        idx = None
        db = self.db
        assert db is not None
        try:
            idx = db.index(item)
        except Exception:
            path = getattr(item, 'path', None)
            if path:
                for i, x in enumerate(db):
                    if getattr(x, 'path', None) == path:
                        idx = i
                        break
        return idx

    def deletion_done(self, job, succeeded=True):
        db_items = self.marked_for_deletion.pop(job, [])
        rows = []
        for item in db_items:
            idx = self.find_item_in_db(item)
            if idx is not None:
                try:
                    rows.append(self.map.index(idx))
                except ValueError:
                    pass

        for row in rows:
            if not succeeded:
                indices = self.row_indices(self.index(row, 0))
                self.dataChanged.emit(indices[0], indices[-1])
        self.count_changed()

    def paths_deleted(self, paths):
        self.map = list(range(len(self.db)))
        self.resort(False)
        self.research(True)
        self.count_changed()

    def is_row_marked_for_deletion(self, row):
        try:
            item = self.db[self.map[row]]
        except IndexError:
            return False

        path = getattr(item, 'path', None)
        for items in self.marked_for_deletion.values():
            for x in items:
                if x is item or (path and path == getattr(x, 'path', None)):
                    return True
        return False

    def clear_ondevice(self, db_ids, to_what=None):
        for data in self.db:
            if data is None:
                continue
            app_id = getattr(data, 'application_id', None)
            if app_id is not None and app_id in db_ids:
                data.in_library = to_what
            self.beginResetModel(), self.endResetModel()

    def flags(self, index):
        if self.is_row_marked_for_deletion(index.row()):
            return Qt.ItemFlag.NoItemFlags
        flags = QAbstractTableModel.flags(self, index)
        if index.isValid():
            cname = self.column_map[index.column()]
            if cname in self.editable and (
                cname != 'collections'
                or (
                    callable(getattr(self.db, 'supports_collections', None))
                    and self.db.supports_collections()
                    and device_prefs['manage_device_metadata'] == 'manual'
                )
            ):
                flags |= Qt.ItemFlag.ItemIsEditable
        return flags

    def search(self, text, reset=True):
        # This should not be here, but since the DeviceBooksModel does not
        # implement count_changed and I am too lazy to fix that, this kludge
        # will have to do
        self.resize_rows.emit()

        if not text or not text.strip():
            self.map = list(range(len(self.db)))
        else:
            try:
                matches = self.search_engine.parse(text)
            except ParseException:
                self.searched.emit(False)
                return

            self.map = []
            for i in range(len(self.db)):
                if i in matches:
                    self.map.append(i)
        self.resort(reset=False)
        if reset:
            self.beginResetModel(), self.endResetModel()
        self.last_search = text
        if self.last_search:
            self.searched.emit(True)
        self.count_changed()

    def research(self, reset=True):
        self.search(self.last_search, reset)

    def sort(self, column, order=..., reset=True):
        if not isinstance(order, Qt.SortOrder):
            order = Qt.SortOrder.AscendingOrder if order else Qt.SortOrder.DescendingOrder
        descending = order != Qt.SortOrder.AscendingOrder
        cname = self.column_map[column]

        def author_key(x):
            try:
                ax = self.db[x].author_sort
                if not ax:
                    raise Exception('')
            except Exception:
                try:
                    ax = authors_to_string(self.db[x].authors)
                except Exception:
                    ax = ''
            try:
                return sort_key(ax)
            except Exception:
                return ax

        keygen = {
            'title': ('title_sorter', lambda x: sort_key(x) if x else b''),
            'authors': author_key,
            'size': ('size', int),
            'timestamp': ('datetime', functools.partial(dt_factory, assume_utc=True)),
            'collections': ('device_collections', lambda x: sorted(x, key=sort_key)),
            'inlibrary': ('in_library', lambda x: x or ''),
        }[cname]
        keygen = keygen if callable(keygen) else DeviceDBSortKeyGen(keygen[0], keygen[1], self.db)
        keygen_fn = cast(Callable[[int], bytes | str], keygen)
        self.map.sort(key=keygen_fn, reverse=descending)
        if len(self.map) == len(self.db):
            self.sorted_map = list(self.map)
        else:
            self.sorted_map = list(range(len(self.db)))
            self.sorted_map.sort(key=keygen_fn, reverse=descending)
        self.sorted_on = (self.column_map[column], not descending)
        self.sort_history.insert(0, self.sorted_on)
        if hasattr(keygen, 'db'):
            setattr(keygen, 'db', None)
        if reset:
            self.beginResetModel(), self.endResetModel()

    def resort(self, reset=True):
        if self.sorted_on:
            self.sort(self.column_map.index(self.sorted_on[0]), self.sorted_on[1], reset=False)
        if reset:
            self.beginResetModel(), self.endResetModel()

    def columnCount(self, parent: QModelIndex = QModelIndex()):
        if parent is not None and parent.isValid():
            return 0
        return len(self.column_map)

    def rowCount(self, parent: QModelIndex = QModelIndex()):
        if parent is not None and parent.isValid():
            return 0
        return len(self.map)

    def set_database(self, db):
        self.custom_columns = {}
        self.zero_page_cache = {}
        self.db = db
        self.map = list(range(len(db)))
        self.research(reset=False)
        self.resort()
        self.count_changed()

    def cover(self, row_number):
        item = self.db[self.map[row_number]]
        cdata = item.thumbnail
        img = QImage()
        if cdata is not None:
            if hasattr(cdata, 'image_path'):
                img.load(cdata.image_path)
            elif cdata:
                if isinstance(cdata, (tuple, list)):
                    load_from_data_without_gil(img, cdata[-1])
                else:
                    load_from_data_without_gil(img, cdata)
        if img.isNull():
            img = self.default_image
        return img

    def get_book_display_info(self, idx):
        from calibre.ebooks.metadata.book.base import Metadata

        item = self.db[self.map[idx]]
        cover = self.cover(idx)
        if cover is self.default_image:
            cover = None
        title = item.title
        if not title:
            title = _('Unknown')
        au = item.authors
        if not au:
            au = [_('Unknown')]
        mi = Metadata(title, au)
        mi.cover_data = ('jpg', cover)
        fmt = _('Unknown')
        ext = os.path.splitext(item.path)[1]
        if ext:
            fmt = ext[1:].lower()
        mi.formats = [fmt]
        mi.path = item.path or None
        dt = dt_factory(item.datetime, assume_utc=True)
        mi.timestamp = dt
        mi.device_collections = list(item.device_collections)
        mi.tags = list(getattr(item, 'tags', []))
        mi.comments = getattr(item, 'comments', None)
        mi.book_display_info_is_from_device = True
        series = getattr(item, 'series', None)
        if series:
            sidx = getattr(item, 'series_index', 0)
            mi.series = series
            mi.series_index = sidx
        return mi

    def current_changed(self, current, previous, emit_signal=True):
        if current.isValid():
            idx = current.row()
            try:
                self.db[self.map[idx]]
            except Exception:
                return  # can happen if the device is ejected
            try:
                data = self.get_book_display_info(idx)
            except Exception:
                import traceback

                error_dialog(
                    None,
                    _('Unhandled error'),
                    _('Failed to read book data from calibre library. Click "Show details" for more information'),
                    det_msg=traceback.format_exc(),
                    show=True,
                )
            else:
                if emit_signal:
                    self.new_bookdisplay_data.emit(data)
                else:
                    return data

    def paths(self, rows):
        return [self.db[self.map[r.row()]].path for r in rows]

    def id(self, row):
        row = getattr(row, 'row', lambda: row)()
        try:
            return self.db[self.map[row]].path
        except Exception:
            return None

    def paths_for_db_ids(self, db_ids, as_map=False):
        res = defaultdict(list) if as_map else []
        for r, b in enumerate(self.db):
            if b.application_id in db_ids:
                if as_map:
                    res[b.application_id].append(b)
                else:
                    assert isinstance(res, list)
                    res.append((r, b))
        return res

    def get_collections_with_ids(self):
        collections = set()
        for book in self.db:
            if book.device_collections is not None:
                collections.update(set(book.device_collections))
        self.collections = []
        result = []
        for i, collection in enumerate(collections):
            result.append((i, collection))
            self.collections.append(collection)
        return result

    def rename_collection(self, old_id, new_name):
        old_name = self.collections[old_id]
        for book in self.db:
            if book.device_collections is None:
                continue
            if old_name in book.device_collections:
                book.device_collections.remove(old_name)
                if new_name not in book.device_collections:
                    book.device_collections.append(new_name)

    def delete_collection_using_id(self, old_id):
        old_name = self.collections[old_id]
        for book in self.db:
            if book.device_collections is None:
                continue
            if old_name in book.device_collections:
                book.device_collections.remove(old_name)

    def indices(self, rows):
        """
        Return indices into underlying database from rows
        """
        return [self.map[r.row()] for r in rows]

    def data(self, index, role=...):
        row, column = index.row(), index.column()
        cname = self.column_map[column]
        if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
            if cname == 'title':
                text = self.db[self.map[row]].title
                if not text:
                    text = self.unknown
                return text
            elif cname == 'authors':
                au = self.db[self.map[row]].authors
                if not au:
                    au = [_('Unknown')]
                return authors_to_string(au)
            elif cname == 'size':
                size = self.db[self.map[row]].size
                if not isinstance(size, numbers.Number):
                    size = 0
                return human_readable(size)
            elif cname == 'timestamp':
                dt = self.db[self.map[row]].datetime
                try:
                    dt = dt_factory(dt, assume_utc=True, as_utc=False)
                except OverflowError:
                    dt = dt_factory(time.gmtime(), assume_utc=True, as_utc=False)
                return strftime(TIME_FMT, dt.timetuple())
            elif cname == 'collections':
                tags = self.db[self.map[row]].device_collections
                if tags:
                    tags.sort(key=sort_key)
                    return ', '.join(tags)
            elif DEBUG and cname == 'inlibrary':
                return self.db[self.map[row]].in_library
        elif role == Qt.ItemDataRole.ToolTipRole and index.isValid():
            if column == 0 and hasattr(self.db[self.map[row]], 'in_library_waiting'):
                return _('Waiting for metadata to be updated')
            if self.is_row_marked_for_deletion(row):
                return _('Marked for deletion')
            if cname in ['title', 'authors'] or (
                cname == 'collections' and (callable(getattr(self.db, 'supports_collections', None)) and self.db.supports_collections())
            ):
                return _('Double click to <b>edit</b> me<br><br>')
        elif role == Qt.ItemDataRole.DecorationRole and cname == 'inlibrary':
            if hasattr(self.db[self.map[row]], 'in_library_waiting'):
                return self.sync_icon
            elif self.db[self.map[row]].in_library:
                return self.bool_yes_icon
            elif self.db[self.map[row]].in_library is not None:
                return self.bool_no_icon
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            cname = self.column_map[index.column()]
            ans = Qt.AlignmentFlag.AlignVCenter | ALIGNMENT_MAP[self.alignment_map.get(cname, 'left')]
            return int(ans)  # https://bugreports.qt.io/browse/PYSIDE-1974
        return None

    def headerData(self, section, orientation, role=...):
        if role == Qt.ItemDataRole.ToolTipRole and orientation == Qt.Orientation.Horizontal:
            cname = self.column_map[section]
            text = self.headers[cname]
            return '<b>{}</b>: {}'.format(
                prepare_string_for_xml(text),
                prepare_string_for_xml(_('The lookup/search name is')) + f' <i>{self.column_map[section]}</i>',
            )
        if DEBUG and role == Qt.ItemDataRole.ToolTipRole and orientation == Qt.Orientation.Vertical:
            return _('This book\'s UUID is "{0}"').format(self.db[self.map[section]].uuid)
        if role != Qt.ItemDataRole.DisplayRole:
            return None
        if orientation == Qt.Orientation.Horizontal:
            cname = self.column_map[section]
            text = self.headers[cname]
            return text
        else:
            return section + 1

    def setData(self, index, value, role=...):
        from calibre.gui2.ui import get_gui

        gui = get_gui()
        if gui is not None and gui.shutting_down:
            return False
        done = False
        if role == Qt.ItemDataRole.EditRole:
            row, column = index.row(), index.column()
            cname = self.column_map[column]
            if cname in ('size', 'timestamp', 'inlibrary'):
                return False
            val = str(value or '').strip()
            idx = self.map[row]
            if cname == 'collections':
                tags = [i.strip() for i in val.split(',')]
                tags = [t for t in tags if t]
                self.db[idx].device_collections = tags
                self.dataChanged.emit(index, index)
                self.upload_collections.emit(self.db)
                return True

            if cname == 'title':
                self.db[idx].title = val
            elif cname == 'authors':
                self.db[idx].authors = string_to_authors(val)
            self.dataChanged.emit(index, index)
            self.booklist_dirtied.emit()
            done = True
        return done

    def set_editable(self, editable):
        # Cannot edit if metadata is sent on connect. Reason: changes will
        # revert to what is in the library on next connect.
        if isinstance(editable, list):
            self.editable = editable
        elif editable:
            self.editable = ['title', 'authors', 'collections']
        else:
            self.editable = []
        if device_prefs['manage_device_metadata'] == 'on_connect':
            self.editable = []


# }}}



#!/usr/bin/env python
# Copyright (C) 2006-2009 Søren Roug, European Environment Agency
#
# This is free software.  You may redistribute it under the terms
# of the Apache license and the GNU General Public License Version
# 2 or at your option any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Contributor(s): Michael Howitz, gocept gmbh & co. kg
#
# $Id: userfield.py 447 2008-07-10 20:01:30Z roug $

"""Class to show and manipulate user fields in odf documents."""

import sys
import zipfile

from odf.namespaces import OFFICENS
from odf.opendocument import load
from odf.text import UserFieldDecl

OUTENCODING = 'utf-8'


# OpenDocument v.1.0 section 6.7.1
VALUE_TYPES = {
    'float': (OFFICENS, 'value'),
    'percentage': (OFFICENS, 'value'),
    'currency': (OFFICENS, 'value'),
    'date': (OFFICENS, 'date-value'),
    'time': (OFFICENS, 'time-value'),
    'boolean': (OFFICENS, 'boolean-value'),
    'string': (OFFICENS, 'string-value'),
}


class UserFields:
    """List, view and manipulate user fields."""

    # these attributes can be a filename or a file like object
    src_file = None
    dest_file = None

    def __init__(self, src=None, dest=None):
        """Constructor

        src ... source document name, file like object or None for stdin
        dest ... destination document name, file like object or None for stdout

        """
        self.src_file = src
        self.dest_file = dest
        self.document = None

    def loaddoc(self):
        if isinstance(self.src_file, (bytes, str)):
            # src_file is a filename, check if it is a zip-file
            if not zipfile.is_zipfile(self.src_file):
                raise TypeError(f'{self.src_file} is no odt file.')
        elif self.src_file is None:
            # use stdin if no file given
            self.src_file = sys.stdin

        self.document = load(self.src_file)

    def savedoc(self):
        # write output
        assert self.document is not None
        if self.dest_file is None:
            # use stdout if no filename given
            self.document.save('-')
        else:
            self.document.save(self.dest_file)

    def list_fields(self):
        """List (extract) all known user-fields.

        Returns list of user-field names.

        """
        return [x[0] for x in self.list_fields_and_values()]

    def list_fields_and_values(self, field_names=None):
        """List (extract) user-fields with type and value.

        field_names ... list of field names to show or None for all.

        Returns list of tuples (<field name>, <field type>, <value>).

        """
        self.loaddoc()
        assert self.document is not None
        found_fields = []
        all_fields = self.document.getElementsByType(UserFieldDecl)
        for f in all_fields:
            value_type = f.getAttribute('valuetype')
            if value_type == 'string':
                value = f.getAttribute('stringvalue')
            else:
                value = f.getAttribute('value')
            field_name = f.getAttribute('name')

            if field_names is None or field_name in field_names:
                found_fields.append((
                    field_name.encode(OUTENCODING),
                    value_type.encode(OUTENCODING),
                    value.encode(OUTENCODING),
                ))
        return found_fields

    def list_values(self, field_names):
        """Extract the contents of given field names from the file.

        field_names ... list of field names

        Returns list of field values.

        """
        return [x[2] for x in self.list_fields_and_values(field_names)]

    def get(self, field_name):
        """Extract the contents of this field from the file.

        Returns field value or None if field does not exist.

        """
        values = self.list_values([field_name])
        if not values:
            return None
        return values[0]

    def get_type_and_value(self, field_name):
        """Extract the type and contents of this field from the file.

        Returns tuple (<type>, <field-value>) or None if field does not exist.

        """
        fields = self.list_fields_and_values([field_name])
        if not fields:
            return None
        field_name, value_type, value = fields[0]
        return value_type, value

    def update(self, data):
        """Set the value of user fields. The field types will be the same.

        data ... dict, with field name as key, field value as value

        Returns None

        """
        self.loaddoc()
        assert self.document is not None
        all_fields = self.document.getElementsByType(UserFieldDecl)
        for f in all_fields:
            field_name = f.getAttribute('name')
            if field_name in data:
                value_type = f.getAttribute('valuetype')
                value = data.get(field_name)
                if value_type == 'string':
                    f.setAttribute('stringvalue', value)
                else:
                    f.setAttribute('value', value)
        self.savedoc()



# License: GPLv3 Copyright: 2009, John Schember <john@nachtimwald.com>

"""
Device driver for IRex Digiatal Reader
"""

from calibre.devices.usbms.driver import USBMS
from calibre.utils.localization import _


class IREXDR1000(USBMS):
    name = 'IRex Digital Reader 1000 Device Interface'
    description = _('Communicate with the IRex Digital Reader 1000 e-book reader.')
    author = 'John Schember'
    supported_platforms = ['windows', 'osx', 'linux']

    # Ordered list of supported formats
    # Be sure these have an entry in calibre.devices.mime
    FORMATS = ['epub', 'mobi', 'prc', 'html', 'pdf', 'djvu', 'txt']

    VENDOR_ID = [0x1E6B]
    PRODUCT_ID = [0x001]
    BCD = [0x322]

    VENDOR_NAME = 'IREX'
    WINDOWS_MAIN_MEM = 'DR1000'

    OSX_MAIN_MEM = 'iRex DR1000 Media'

    MAIN_MEMORY_VOLUME_LABEL = 'IRex Digital Reader 1000 Main Memory'

    EBOOK_DIR_MAIN = ''
    DELETE_EXTS = ['.mbp']
    SUPPORTS_SUB_DIRS = True


class IREXDR800(IREXDR1000):
    name = 'IRex Digital Reader 800 Device Interface'
    description = _('Communicate with the IRex Digital Reader 800')
    PRODUCT_ID = [0x002]
    WINDOWS_MAIN_MEM = 'DR800'
    FORMATS = ['epub', 'pdb', 'html', 'pdf', 'txt']

    EBOOK_DIR_MAIN = ''
    DELETE_EXTS = []
    SUPPORTS_SUB_DIRS = True



#!/usr/bin/env python
# License: GPLv3 Copyright: 2013, Kovid Goyal <kovid at kovidgoyal.net>

from calibre.customize.conversion import InputFormatPlugin, OptionRecommendation
from calibre.utils.localization import _


class DOCXInput(InputFormatPlugin):
    name = 'DOCX Input'
    author = 'Kovid Goyal'
    description = _('Convert DOCX files (.docx and .docm) to HTML')
    file_types = {'docx', 'docm'}
    commit_name = 'docx_input'

    options = {
        OptionRecommendation(
            name='docx_no_cover',
            recommended_value=False,
            help=_(
                'Normally, if a large image is present at the start of the document that looks like a cover, '
                'it will be removed from the document and used as the cover for created e-book. This option '
                'turns off that behavior.'
            ),
        ),
        OptionRecommendation(
            name='docx_no_pagebreaks_between_notes',
            recommended_value=False,
            help=_('Do not insert a page break after every endnote.'),
        ),
        OptionRecommendation(
            name='docx_inline_subsup',
            recommended_value=False,
            help=_('Render superscripts and subscripts so that they do not affect the line height'),
        ),
    }

    recommendations = {('page_breaks_before', '/', OptionRecommendation.MED)}

    def convert(self, stream, options, file_ext, log, accelerators):
        from calibre.ebooks.docx.to_html import Convert

        c = Convert(
            stream,
            detect_cover=not options.docx_no_cover,
            log=log,
            notes_nopb=options.docx_no_pagebreaks_between_notes,
            nosupsub=options.docx_inline_subsup,
        )
        try:
            return c()
        finally:
            c.docx.close()



#!/usr/bin/env python
# License: GPLv3 Copyright: 2013, Kovid Goyal <kovid at kovidgoyal.net>

from numbers import Number

from calibre.ebooks.metadata.book import ALL_METADATA_FIELDS, TOP_LEVEL_IDENTIFIERS
from calibre.utils.formatter import TemplateFormatter
from calibre.utils.localization import _


class SafeFormat(TemplateFormatter):
    def __init__(self):
        TemplateFormatter.__init__(self)

    def get_value(self, key, args, kwargs):
        if not key or isinstance(key, Number):
            return ''
        orig_key = key = key.lower()
        if key != 'title_sort' and key not in TOP_LEVEL_IDENTIFIERS and key not in ALL_METADATA_FIELDS:
            from calibre.ebooks.metadata.book.base import field_metadata

            key = field_metadata.search_term_to_field_key(key)
            if key is None or (self.book and key not in self.book.all_field_keys()):
                if hasattr(self.book, orig_key):
                    key = orig_key
                else:
                    raise ValueError(_('Value: unknown field ') + orig_key)
        if self.book is None:
            return ''
        try:
            b = self.book.get_user_metadata(key, False)
        except Exception:
            b = None
        if b and b['datatype'] in {'int', 'float'} and self.book.get(key, None) is None:
            v = ''
        else:
            v = self.book.format_field(key, series_with_index=False)[1]
        if v is None:
            return ''
        return v
